import json

import asynctest
from asynctest import TestCase as AsyncTestCase
from asynctest import mock as async_mock

from aries_cloudagent.config.injection_context import InjectionContext
from aries_cloudagent.messaging.request_context import RequestContext

from .....admin.request_context import AdminRequestContext
from .. import routes as test_module
from ..manager import MediationManager
from ..models.mediation_record import MediationRecord


class TestCoordinateMediationRoutes(AsyncTestCase):
    def setUp(self):
        self.session_inject = {}
        self.context = AdminRequestContext.test_context(self.session_inject)
        self.outbound_message_router = async_mock.CoroutineMock()
        self.request_dict = {
            "context": self.context,
            "outbound_message_router": self.outbound_message_router,
        }
        self.request = async_mock.MagicMock(
            match_info={
                "mediation_id": "test-mediation-id",
                "conn_id": "test-conn-id",
            },
            query={},
            json=async_mock.CoroutineMock(return_value={}),
            __getitem__=lambda _, k: self.request_dict[k],
        )
        serialized = {
            "mediation_id": "fake_id",
            "state": "granted",
            "role": "server",
            "connection_id": "c3dd00cf-f6a2-4ddf-93d8-49ae74bdacef",
            "mediator_terms": [],
            "recipient_terms": [],
            "routing_keys": ["EwUKjVLboiLSuoWSEtDvrgrd41EUxG5bLecQrkHB63Up"],
            "endpoint": "http://192.168.1.13:3005",
            "created_at": "1234567890",
        }
        self.mock_record = async_mock.MagicMock(
            **serialized,
            serialize=async_mock.MagicMock(return_value=serialized),
            save=async_mock.CoroutineMock()
        )

    def test_mediation_sort_key(self):
        assert (
            test_module.mediation_sort_key(
                {"state": MediationRecord.STATE_DENIED, "created_at": ""}
            )
            == "2"
        )
        assert (
            test_module.mediation_sort_key(
                {"state": MediationRecord.STATE_REQUEST, "created_at": ""}
            )
            == "1"
        )
        assert (
            test_module.mediation_sort_key(
                {"state": MediationRecord.STATE_GRANTED, "created_at": ""}
            )
            == "0"
        )

    async def test_list_mediation_requests(self):
        self.request.query = {}
        self.context.session = async_mock.CoroutineMock()
        with async_mock.patch.object(
            test_module.MediationRecord,
            "query",
            async_mock.CoroutineMock(return_value=[self.mock_record]),
        ) as mock_query, async_mock.patch.object(
            test_module.web, "json_response"
        ) as json_response:
            await test_module.list_mediation_requests(self.request)
            json_response.assert_called_once_with(
                [self.mock_record.serialize.return_value]
            )
            mock_query.assert_called_once_with(self.context.session.return_value, {})

    async def test_list_mediation_requests_filters(self):
        self.request.query = {
            "state": MediationRecord.STATE_GRANTED,
            "conn_id": "test-conn-id",
        }
        self.context.session = async_mock.CoroutineMock()
        with async_mock.patch.object(
            test_module.MediationRecord,
            "query",
            async_mock.CoroutineMock(return_value=[self.mock_record]),
        ) as mock_query, async_mock.patch.object(
            test_module.web, "json_response"
        ) as json_response:
            await test_module.list_mediation_requests(self.request)
            json_response.assert_called_once_with(
                [self.mock_record.serialize.return_value]
            )
            mock_query.assert_called_once_with(
                self.context.session.return_value,
                {
                    "connection_id": "test-conn-id",
                    "state": MediationRecord.STATE_GRANTED,
                },
            )

    async def test_list_mediation_requests_x(self):
        with async_mock.patch.object(
            test_module,
            "MediationRecord",
            async_mock.MagicMock(
                query=async_mock.CoroutineMock(side_effect=test_module.StorageError())
            ),
        ) as mock_med_rec:
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.list_mediation_requests(self.request)

    async def test_list_mediation_requests_no_records(self):
        with async_mock.patch.object(
            test_module,
            "MediationRecord",
            async_mock.MagicMock(query=async_mock.CoroutineMock(return_value=[])),
        ) as mock_med_rec, async_mock.patch.object(
            test_module.web, "json_response"
        ) as mock_response:
            await test_module.list_mediation_requests(self.request)
            mock_response.assert_called_once_with([])

    async def test_retrieve_mediation_request(self):
        with async_mock.patch.object(
            test_module.MediationRecord, "retrieve_by_id", async_mock.CoroutineMock()
        ) as mock_mediation_record_retrieve, async_mock.patch.object(
            test_module.web, "json_response"
        ) as mock_response:
            mock_mediation_record_retrieve.return_value = self.mock_record
            await test_module.retrieve_mediation_request(self.request)
            mock_response.assert_called_once_with(
                self.mock_record.serialize.return_value
            )
            mock_mediation_record_retrieve.assert_called()

    async def test_retrieve_mediation_request_x_not_found(self):
        with async_mock.patch.object(
            test_module.MediationRecord,
            "retrieve_by_id",
            async_mock.CoroutineMock(side_effect=test_module.StorageNotFoundError()),
        ) as mock_mediation_record_retrieve, async_mock.patch.object(
            test_module.web, "json_response"
        ) as mock_response, self.assertRaises(
            test_module.web.HTTPNotFound
        ):
            await test_module.retrieve_mediation_request(self.request)

    async def test_retrieve_mediation_request_x_storage_error(self):
        with async_mock.patch.object(
            test_module.MediationRecord,
            "retrieve_by_id",
            async_mock.CoroutineMock(side_effect=test_module.StorageError()),
        ) as mock_mediation_record_retrieve, async_mock.patch.object(
            test_module.web, "json_response"
        ) as mock_response, self.assertRaises(
            test_module.web.HTTPBadRequest
        ):
            await test_module.retrieve_mediation_request(self.request)

    async def test_delete_mediation_request(self):
        with async_mock.patch.object(
            test_module.MediationRecord, "retrieve_by_id", async_mock.CoroutineMock()
        ) as mock_mediation_record_retrieve, async_mock.patch.object(
            self.mock_record, "delete_record", async_mock.CoroutineMock()
        ) as mock_delete_record, async_mock.patch.object(
            test_module.web, "json_response"
        ) as mock_response:
            mock_mediation_record_retrieve.return_value = self.mock_record
            await test_module.delete_mediation_request(self.request)
            mock_response.assert_called_once_with(
                self.mock_record.serialize.return_value
            )
            mock_mediation_record_retrieve.assert_called()
            mock_delete_record.assert_called()

    async def test_delete_mediation_request_x_not_found(self):
        with async_mock.patch.object(
            test_module.MediationRecord,
            "retrieve_by_id",
            async_mock.CoroutineMock(side_effect=test_module.StorageNotFoundError()),
        ) as mock_mediation_record_retrieve, async_mock.patch.object(
            test_module.web, "json_response"
        ) as mock_response, self.assertRaises(
            test_module.web.HTTPNotFound
        ):
            await test_module.delete_mediation_request(self.request)

    async def test_delete_mediation_request_x_storage_error(self):
        with async_mock.patch.object(
            test_module.MediationRecord,
            "retrieve_by_id",
            async_mock.CoroutineMock(side_effect=test_module.StorageError()),
        ) as mock_mediation_record_retrieve, async_mock.patch.object(
            test_module.web, "json_response"
        ) as mock_response, self.assertRaises(
            test_module.web.HTTPBadRequest
        ):
            await test_module.delete_mediation_request(self.request)

    async def test_request_mediation(self):
        body = {
            "mediator_terms": ["meaningless string because terms are not used"],
            "recipient_terms": ["meaningless string because terms are not a 'thing'"],
        }
        self.request.json.return_value = body
        with async_mock.patch.object(
            test_module, "MediationManager", autospec=True
        ) as mock_med_mgr, async_mock.patch.object(
            test_module.web, "json_response"
        ) as mock_response, async_mock.patch.object(
            test_module.MediationRecord,
            "exists_for_connection_id",
            async_mock.CoroutineMock(return_value=False),
        ) as mock_mediation_record_exists, async_mock.patch.object(
            test_module.ConnRecord, "retrieve_by_id", async_mock.CoroutineMock()
        ) as mock_conn_rec_retrieve_by_id:
            mock_med_mgr.return_value.prepare_request = async_mock.CoroutineMock(
                return_value=(
                    self.mock_record,
                    async_mock.MagicMock(  # mediation request
                        serialize=async_mock.MagicMock(return_value={"a": "value"}),
                    ),
                )
            )
            await test_module.request_mediation(self.request)
            mock_response.assert_called_once_with(
                self.mock_record.serialize.return_value, status=201
            )
            self.outbound_message_router.assert_called()

    async def test_request_mediation_x_conn_not_ready(self):
        body = {
            "mediator_terms": ["meaningless string because terms are not used"],
            "recipient_terms": ["meaningless string because terms are not a 'thing'"],
        }
        self.request.json.return_value = body
        with async_mock.patch.object(
            test_module.ConnRecord,
            "retrieve_by_id",
            async_mock.CoroutineMock(return_value=async_mock.MagicMock(is_ready=False)),
        ) as mock_conn_rec_retrieve_by_id, self.assertRaises(
            test_module.web.HTTPBadRequest
        ) as exc:
            await test_module.request_mediation(self.request)
            assert "request connection is not ready" in exc.msg

    async def test_request_mediation_x_already_exists(self):
        body = {
            "mediator_terms": ["meaningless string because terms are not used"],
            "recipient_terms": ["meaningless string because terms are not a 'thing'"],
        }
        self.request.json.return_value = body
        with async_mock.patch.object(
            test_module.ConnRecord, "retrieve_by_id", async_mock.CoroutineMock()
        ) as mock_conn_rec_retrieve_by_id, async_mock.patch.object(
            test_module.MediationRecord,
            "exists_for_connection_id",
            async_mock.CoroutineMock(return_value=True),
        ) as mock_exists_for_conn, self.assertRaises(
            test_module.web.HTTPBadRequest
        ) as exc:
            await test_module.request_mediation(self.request)
            assert "already exists for connection" in exc.msg

    async def test_request_mediation_x_conn_not_found(self):
        body = {
            "mediator_terms": ["meaningless string because terms are not used"],
            "recipient_terms": ["meaningless string because terms are not a 'thing'"],
        }
        self.request.json.return_value = body
        with async_mock.patch.object(
            test_module.ConnRecord,
            "retrieve_by_id",
            async_mock.CoroutineMock(side_effect=test_module.StorageNotFoundError()),
        ) as mock_conn_rec_retrieve_by_id, self.assertRaises(
            test_module.web.HTTPNotFound
        ):
            await test_module.request_mediation(self.request)

    async def test_request_mediation_x_storage_error(self):
        body = {
            "mediator_terms": ["meaningless string because terms are not used"],
            "recipient_terms": ["meaningless string because terms are not a 'thing'"],
        }
        self.request.json.return_value = body
        with async_mock.patch.object(
            test_module.ConnRecord,
            "retrieve_by_id",
            async_mock.CoroutineMock(side_effect=test_module.StorageError()),
        ) as mock_conn_rec_retrieve_by_id, self.assertRaises(
            test_module.web.HTTPBadRequest
        ):
            await test_module.request_mediation(self.request)

    async def test_mediation_request_grant_role_server(self):
        self.mock_record.role = MediationRecord.ROLE_SERVER
        with async_mock.patch.object(
            test_module.MediationRecord,
            "retrieve_by_id",
            async_mock.CoroutineMock(return_value=self.mock_record),
        ) as mock_mediation_record_retrieve, async_mock.patch.object(
            test_module.web, "json_response"
        ) as mock_response:
            await test_module.mediation_request_grant(self.request)
            mock_response.assert_called_once_with(
                self.mock_record.serialize.return_value, status=201
            )
            self.outbound_message_router.assert_called()

    async def test_mediation_request_grant_role_client_x(self):
        self.mock_record.role = MediationRecord.ROLE_CLIENT
        with async_mock.patch.object(
            test_module.MediationRecord,
            "retrieve_by_id",
            async_mock.CoroutineMock(return_value=self.mock_record),
        ), self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.mediation_request_grant(self.request)

    async def test_mediation_request_grant_x_rec_not_found(self):
        with async_mock.patch.object(
            test_module.MediationRecord,
            "retrieve_by_id",
            async_mock.CoroutineMock(side_effect=test_module.StorageNotFoundError()),
        ), self.assertRaises(test_module.web.HTTPNotFound):
            await test_module.mediation_request_grant(self.request)

    async def test_mediation_request_grant_x_storage_error(self):
        with async_mock.patch.object(
            test_module.MediationRecord,
            "retrieve_by_id",
            async_mock.CoroutineMock(side_effect=test_module.StorageError()),
        ), self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.mediation_request_grant(self.request)

    async def test_mediation_request_deny_role_server(self):
        self.mock_record.role = MediationRecord.ROLE_SERVER
        with async_mock.patch.object(
            test_module.MediationRecord,
            "retrieve_by_id",
            async_mock.CoroutineMock(return_value=self.mock_record),
        ) as mock_mediation_record_retrieve, async_mock.patch.object(
            test_module.web, "json_response"
        ) as mock_response:
            await test_module.mediation_request_deny(self.request)
            mock_response.assert_called_once_with(
                self.mock_record.serialize.return_value, status=201
            )
            self.outbound_message_router.assert_called()

    async def test_mediation_request_deny_role_client_x(self):
        self.mock_record.role = MediationRecord.ROLE_CLIENT
        with async_mock.patch.object(
            test_module.MediationRecord, "retrieve_by_id", async_mock.CoroutineMock()
        ) as mock_mediation_record_retrieve, async_mock.patch.object(
            test_module.web, "json_response"
        ):
            mock_mediation_record_retrieve.return_value = async_mock.MagicMock(
                role=MediationRecord.ROLE_CLIENT
            )
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.mediation_request_deny(self.request)

    async def test_mediation_request_deny_x_rec_not_found(self):
        with async_mock.patch.object(
            test_module.MediationRecord,
            "retrieve_by_id",
            async_mock.CoroutineMock(side_effect=test_module.StorageNotFoundError()),
        ), self.assertRaises(test_module.web.HTTPNotFound):
            await test_module.mediation_request_deny(self.request)

    async def test_mediation_request_deny_x_storage_error(self):
        with async_mock.patch.object(
            test_module.MediationRecord,
            "retrieve_by_id",
            async_mock.CoroutineMock(side_effect=test_module.StorageError()),
        ), self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.mediation_request_deny(self.request)

    async def test_get_keylist(self):
        self.request.query["role"] = MediationRecord.ROLE_SERVER
        self.request.query["conn_id"] = "test-id"

        query_results = [
            async_mock.MagicMock(
                serialize=async_mock.MagicMock(
                    return_value={"serialized": "route record"}
                )
            )
        ]

        with async_mock.patch.object(
            test_module.RouteRecord,
            "query",
            async_mock.CoroutineMock(return_value=query_results),
        ) as mock_query, async_mock.patch.object(
            self.context,
            "session",
            async_mock.MagicMock(return_value=self.context.session()),
        ) as mock_session, async_mock.patch.object(
            test_module.web, "json_response"
        ) as mock_response:
            await test_module.get_keylist(self.request)
            mock_response.assert_called_once_with(
                [{"serialized": "route record"}], status=200
            )
            mock_query.assert_called_once_with(
                mock_session.return_value,
                {"role": MediationRecord.ROLE_SERVER, "connection_id": "test-id"},
            )

    async def test_get_keylist_no_matching_records(self):
        with async_mock.patch.object(
            test_module.RouteRecord,
            "query",
            async_mock.CoroutineMock(return_value=[]),
        ) as mock_query, async_mock.patch.object(
            self.context,
            "session",
            async_mock.MagicMock(return_value=self.context.session()),
        ) as mock_session, async_mock.patch.object(
            test_module.web, "json_response"
        ) as mock_response:
            await test_module.get_keylist(self.request)
            mock_query.assert_called_once_with(mock_session.return_value, {})
            mock_response.assert_called_once_with([], status=200)

    async def test_get_keylist_storage_error(self):
        with async_mock.patch.object(
            test_module.RouteRecord,
            "query",
            async_mock.CoroutineMock(side_effect=test_module.StorageError),
        ) as mock_query, self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.get_keylist(self.request)

    async def test_send_keylist_update(self):
        body = {
            "updates": [
                {"recipient_key": "test-key0", "action": "add"},
                {"recipient_key": "test-key1", "action": "remove"},
            ]
        }

        self.request.json.return_value = body

        with async_mock.patch.object(
            test_module.MediationRecord,
            "retrieve_by_id",
            async_mock.CoroutineMock(
                return_value=async_mock.MagicMock(
                    state=MediationRecord.STATE_GRANTED, connection_id="test-conn-id"
                )
            ),
        ) as mock_retrieve_by_id, async_mock.patch.object(
            test_module.web,
            "json_response",
            async_mock.MagicMock(
                side_effect=lambda *args, **kwargs: [*args, *kwargs.values()]
            ),
        ) as mock_response:
            results, status = await test_module.send_keylist_update(self.request)
            assert results["updates"] == body["updates"]
            assert status == 201

    async def test_send_keylist_update_bad_action(self):
        self.request.json.return_value = {
            "updates": [
                {"recipient_key": "test-key0", "action": "wrong"},
            ]
        }

        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.send_keylist_update(self.request)

    async def test_send_keylist_update_bad_mediation_state(self):
        self.request.json.return_value = {
            "updates": [
                {"recipient_key": "test-key0", "action": "add"},
            ]
        }

        with async_mock.patch.object(
            test_module.MediationRecord,
            "retrieve_by_id",
            async_mock.CoroutineMock(
                return_value=async_mock.MagicMock(
                    state=MediationRecord.STATE_DENIED, connection_id="test-conn-id"
                )
            ),
        ) as mock_retrieve_by_id, self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.send_keylist_update(self.request)

    async def test_send_keylist_update_bad_updates(self):
        self.request.json.return_value = {"updates": []}
        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.send_keylist_update(self.request)

    async def test_send_keylist_update_x_no_mediation_rec(self):
        self.request.json.return_value = {
            "updates": [
                {"recipient_key": "test-key0", "action": "add"},
            ]
        }
        with async_mock.patch.object(
            test_module.MediationRecord,
            "retrieve_by_id",
            async_mock.CoroutineMock(side_effect=test_module.StorageNotFoundError()),
        ), self.assertRaises(test_module.web.HTTPNotFound):
            await test_module.send_keylist_update(self.request)

    async def test_send_keylist_update_x_storage_error(self):
        self.request.json.return_value = {
            "updates": [
                {"recipient_key": "test-key0", "action": "add"},
            ]
        }

        with async_mock.patch.object(
            test_module.MediationRecord,
            "retrieve_by_id",
            async_mock.CoroutineMock(side_effect=test_module.StorageError()),
        ), self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.send_keylist_update(self.request)

    @async_mock.patch.object(test_module, "MediationManager", autospec=True)
    async def test_send_keylist_query(self, mock_manager):
        self.request.json.return_value = {"filter": {"test": "filter"}}
        self.request.query = {"paginate_limit": 10, "paginate_offset": 20}
        with async_mock.patch.object(
            test_module.MediationRecord,
            "retrieve_by_id",
            async_mock.CoroutineMock(return_value=self.mock_record),
        ) as mock_retrieve_by_id, async_mock.patch.object(
            mock_manager.return_value,
            "prepare_keylist_query",
            async_mock.CoroutineMock(),
        ) as mock_prepare_keylist_query, async_mock.patch.object(
            test_module.web, "json_response"
        ) as mock_response:
            await test_module.send_keylist_query(self.request)
            mock_prepare_keylist_query.assert_called_once_with(
                filter_={"test": "filter"}, paginate_limit=10, paginate_offset=20
            )
            self.outbound_message_router.assert_called()
            mock_response.assert_called_once_with(
                mock_prepare_keylist_query.return_value.serialize.return_value,
                status=201,
            )

    async def test_send_keylist_query_x_no_mediation_record(self):
        with async_mock.patch.object(
            test_module.MediationRecord,
            "retrieve_by_id",
            async_mock.CoroutineMock(side_effect=test_module.StorageNotFoundError()),
        ) as mock_retrieve_by_id, self.assertRaises(test_module.web.HTTPNotFound):
            await test_module.send_keylist_query(self.request)

    async def test_send_keylist_query_x_storage_error(self):
        with async_mock.patch.object(
            test_module.MediationRecord,
            "retrieve_by_id",
            async_mock.CoroutineMock(side_effect=test_module.StorageError()),
        ) as mock_retrieve_by_id, self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.send_keylist_query(self.request)

    async def test_get_default_mediator(self):
        self.request.query = {}
        self.context.session = async_mock.CoroutineMock()
        with async_mock.patch.object(
            test_module.web, "json_response"
        ) as json_response, async_mock.patch.object(
            test_module.MediationManager,
            "get_default_mediator",
            async_mock.CoroutineMock(return_value=self.mock_record),
        ) as mock_mgr_get_default_record:
            await test_module.get_default_mediator(self.request)
            json_response.assert_called_once_with(
                self.mock_record.serialize.return_value,
                status=200,
            )

    async def test_get_empty_default_mediator(self):
        self.request.query = {}
        self.context.session = async_mock.CoroutineMock()
        with async_mock.patch.object(
            test_module.web, "json_response"
        ) as json_response, async_mock.patch.object(
            test_module.MediationManager,
            "get_default_mediator",
            async_mock.CoroutineMock(return_value=None),
        ) as mock_mgr_get_default_record:
            await test_module.get_default_mediator(self.request)
            json_response.assert_called_once_with(
                {},
                status=200,
            )

    async def test_get_default_mediator_storage_error(self):
        self.request.query = {}
        self.context.session = async_mock.CoroutineMock()
        with async_mock.patch.object(
            test_module.web, "json_response"
        ) as json_response, async_mock.patch.object(
            test_module.MediationManager,
            "get_default_mediator",
            async_mock.CoroutineMock(side_effect=test_module.StorageNotFoundError()),
        ) as mock_mgr_get_default_record:
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.get_default_mediator(self.request)

    async def test_set_default_mediator(self):
        self.request.match_info = {
            "mediation_id": "fake_id",
        }
        self.request.query = {}
        self.context.session = async_mock.CoroutineMock()
        with async_mock.patch.object(
            test_module.MediationManager,
            "get_default_mediator",
            async_mock.CoroutineMock(return_value=self.mock_record),
        ) as mock_mgr_get_default_record, async_mock.patch.object(
            test_module.MediationManager,
            "set_default_mediator_by_id",
            async_mock.CoroutineMock(),
        ) as mock_mgr_set_default_record_by_id, async_mock.patch.object(
            test_module.web, "json_response"
        ) as json_response:
            await test_module.set_default_mediator(self.request)
            json_response.assert_called_once_with(
                self.mock_record.serialize.return_value,
                status=201,
            )

    async def test_set_default_mediator_storage_error(self):
        self.request.match_info = {
            "mediation_id": "bad_id",
        }
        self.request.query = {}
        self.context.session = async_mock.CoroutineMock()
        with async_mock.patch.object(
            test_module.MediationManager,
            "get_default_mediator",
            async_mock.CoroutineMock(side_effect=test_module.StorageError()),
        ) as mock_mgr_get_default_record, async_mock.patch.object(
            test_module.MediationManager,
            "set_default_mediator_by_id",
            async_mock.CoroutineMock(side_effect=test_module.StorageError()),
        ) as mock_mgr_set_default_record_by_id, async_mock.patch.object(
            test_module.web, "json_response"
        ) as json_response:
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.set_default_mediator(self.request)

    async def test_clear_default_mediator(self):
        self.request.query = {}
        self.context.session = async_mock.CoroutineMock()
        with async_mock.patch.object(
            test_module.MediationManager,
            "get_default_mediator",
            async_mock.CoroutineMock(return_value=self.mock_record),
        ) as mock_mgr_get_default_record, async_mock.patch.object(
            test_module.MediationManager,
            "clear_default_mediator",
            async_mock.CoroutineMock(),
        ) as mock_mgr_clear_default_record_by_id, async_mock.patch.object(
            test_module.web, "json_response"
        ) as json_response:
            await test_module.clear_default_mediator(self.request)
            json_response.assert_called_once_with(
                self.mock_record.serialize.return_value,
                status=201,
            )

    async def test_clear_default_mediator_storage_error(self):
        self.request.query = {}
        self.context.session = async_mock.CoroutineMock()
        with async_mock.patch.object(
            test_module.MediationManager,
            "get_default_mediator",
            async_mock.CoroutineMock(side_effect=test_module.StorageError()),
        ) as mock_mgr_get_default_record, async_mock.patch.object(
            test_module.MediationManager,
            "clear_default_mediator",
            async_mock.CoroutineMock(),
        ) as mock_mgr_clear_default_record_by_id, async_mock.patch.object(
            test_module.web, "json_response"
        ) as json_response:
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.clear_default_mediator(self.request)

    async def test_register(self):
        mock_app = async_mock.MagicMock()
        mock_app.add_routes = async_mock.MagicMock()

        await test_module.register(mock_app)
        mock_app.add_routes.assert_called_once()

    async def test_post_process_routes(self):
        mock_app = async_mock.MagicMock(_state={"swagger_dict": {}})
        test_module.post_process_routes(mock_app)
        assert "tags" in mock_app._state["swagger_dict"]



import pytest

from asynctest import mock as async_mock

from ....config.injection_context import InjectionContext
from ....core.error import ProfileError
from ....ledger.indy import IndySdkLedgerPool

from ..profile import IndySdkProfile
from ..wallet_setup import IndyWalletConfig, IndyOpenWallet


@pytest.fixture()
async def profile():
    context = InjectionContext()
    context.injector.bind_instance(IndySdkLedgerPool, IndySdkLedgerPool("name"))
    yield IndySdkProfile(
        IndyOpenWallet(
            config=IndyWalletConfig({"name": "test-profile"}),
            created=True,
            handle=1,
            master_secret_id="master-secret",
        ),
        context,
    )


class TestIndySdkProfile:
    @pytest.mark.asyncio
    async def test_properties(self, profile):
        assert profile.name == "test-profile"
        assert profile.backend == "indy"
        assert profile.wallet and profile.wallet.handle == 1

        assert "IndySdkProfile" in str(profile)
        assert profile.created
        assert profile.wallet.created
        assert profile.wallet.master_secret_id == "master-secret"

        with async_mock.patch.object(profile, "opened", False):
            with pytest.raises(ProfileError):
                await profile.remove()

        with async_mock.patch.object(
            profile.opened, "close", async_mock.CoroutineMock()
        ):
            await profile.remove()
            assert profile.opened is None

    def test_read_only(self):
        context = InjectionContext(settings={"ledger.read_only": True})
        context.injector.bind_instance(IndySdkLedgerPool, IndySdkLedgerPool("name"))
        ro_profile = IndySdkProfile(
            IndyOpenWallet(
                config=IndyWalletConfig({"name": "test-profile"}),
                created=True,
                handle=1,
                master_secret_id="master-secret",
            ),
            context,
        )



"""coordinate mediation admin routes."""

# https://github.com/hyperledger/aries-rfcs/tree/master/features/0211-route-coordination#0211-mediator-coordination-protocol

from aiohttp import web
from aiohttp_apispec import (
    docs,
    match_info_schema,
    querystring_schema,
    request_schema,
    response_schema,
)
from marshmallow import fields, validate

from ....admin.request_context import AdminRequestContext
from ....connections.models.conn_record import ConnRecord
from ....messaging.models.base import BaseModelError
from ....messaging.models.openapi import OpenAPISchema
from ....messaging.valid import UUIDFour
from ....storage.error import StorageError, StorageNotFoundError
from ...connections.v1_0.routes import ConnIdMatchInfoSchema
from ...routing.v1_0.models.route_record import RouteRecord, RouteRecordSchema
from .manager import MediationManager, MediationManagerError
from .message_types import SPEC_URI
from .messages.inner.keylist_update_rule import (
    KeylistUpdateRule,
    KeylistUpdateRuleSchema,
)
from .messages.keylist_query import KeylistQuerySchema
from .messages.keylist_update import KeylistUpdateSchema
from .messages.mediate_deny import MediationDenySchema
from .messages.mediate_grant import MediationGrantSchema
from .models.mediation_record import MediationRecord, MediationRecordSchema


CONNECTION_ID_SCHEMA = fields.UUID(
    description="Connection identifier (optional)",
    required=False,
    example=UUIDFour.EXAMPLE,
)


MEDIATION_ID_SCHEMA = fields.UUID(
    description="Mediation record identifier",
    example=UUIDFour.EXAMPLE,
)


MEDIATION_STATE_SCHEMA = fields.Str(
    description="Mediation state (optional)",
    required=False,
    validate=validate.OneOf(
        [
            getattr(MediationRecord, m)
            for m in vars(MediationRecord)
            if m.startswith("STATE_")
        ]
    ),
    example="request, granted, or denied",
)


MEDIATOR_TERMS_SCHEMA = fields.List(
    fields.Str(
        description="Indicate terms that the mediator "
        "requires the recipient to agree to"
    ),
    required=False,
    description="List of mediator rules for recipient",
)


RECIPIENT_TERMS_SCHEMA = fields.List(
    fields.Str(
        description="Indicate terms that the recipient "
        "requires the mediator to agree to"
    ),
    required=False,
    description="List of recipient rules for mediation",
)


class MediationListSchema(OpenAPISchema):
    """Result schema for mediation list query."""

    results = fields.List(
        fields.Nested(MediationRecordSchema),
        description="List of mediation records",
    )


class MediationListQueryStringSchema(OpenAPISchema):
    """Parameters and validators for mediation record list request query string."""

    conn_id = CONNECTION_ID_SCHEMA
    mediator_terms = MEDIATOR_TERMS_SCHEMA
    recipient_terms = RECIPIENT_TERMS_SCHEMA
    state = MEDIATION_STATE_SCHEMA


class MediationCreateRequestSchema(OpenAPISchema):
    """Parameters and validators for create Mediation request query string."""

    mediator_terms = MEDIATOR_TERMS_SCHEMA
    recipient_terms = RECIPIENT_TERMS_SCHEMA


class AdminMediationDenySchema(OpenAPISchema):
    """Parameters and validators for Mediation deny admin request query string."""

    mediator_terms = MEDIATOR_TERMS_SCHEMA
    recipient_terms = RECIPIENT_TERMS_SCHEMA


class MediationIdMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking mediation request id."""

    mediation_id = MEDIATION_ID_SCHEMA


class GetKeylistQuerySchema(OpenAPISchema):
    """Get keylist query string paramaters."""

    conn_id = CONNECTION_ID_SCHEMA
    role = fields.Str(
        description=f"Filer on role, '{MediationRecord.ROLE_CLIENT}' for keys \
        mediated by other agents, '{MediationRecord.ROLE_SERVER}' for keys \
        mediated by this agent",
        validate=validate.OneOf(
            [MediationRecord.ROLE_CLIENT, MediationRecord.ROLE_SERVER]
        ),
        missing=MediationRecord.ROLE_SERVER,
        required=False,
    )


class KeylistSchema(OpenAPISchema):
    """Result schema for mediation list query."""

    results = fields.List(
        fields.Nested(RouteRecordSchema),
        description="List of keylist records",
    )


class KeylistQueryFilterRequestSchema(OpenAPISchema):
    """Request schema for keylist query filtering."""

    filter = fields.Dict(required=False, description="Filter for keylist query.")


class KeylistQueryPaginateQuerySchema(OpenAPISchema):
    """Query string schema for keylist query pagination."""

    paginate_limit = fields.Int(
        required=False, missing=-1, description="limit number of results"
    )
    paginate_offset = fields.Int(
        required=False, missing=0, description="offset to use in pagination"
    )


class KeylistUpdateRequestSchema(OpenAPISchema):
    """keylist update request schema."""

    updates = fields.List(fields.Nested(KeylistUpdateRuleSchema()))


def mediation_sort_key(mediation: dict):
    """Get the sorting key for a particular serialized mediation record."""
    if mediation["state"] == MediationRecord.STATE_DENIED:
        pfx = "2"
    elif mediation["state"] == MediationRecord.STATE_REQUEST:
        pfx = "1"
    else:  # GRANTED
        pfx = "0"
    return pfx + mediation["created_at"]


@docs(
    tags=["mediation"],
    summary="Query mediation requests, returns list of all mediation records",
)
@querystring_schema(MediationListQueryStringSchema())
@response_schema(MediationListSchema(), 200)
async def list_mediation_requests(request: web.BaseRequest):
    """List mediation requests for either client or server role."""
    context: AdminRequestContext = request["context"]
    conn_id = request.query.get("conn_id")
    state = request.query.get("state")

    tag_filter = {}
    if conn_id:
        tag_filter["connection_id"] = conn_id
    if state:
        tag_filter["state"] = state

    try:
        session = await context.session()
        records = await MediationRecord.query(session, tag_filter)
        results = [record.serialize() for record in records]
        results.sort(key=mediation_sort_key)
    except (StorageError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err
    return web.json_response(results)


@docs(tags=["mediation"], summary="Retrieve mediation request record")
@match_info_schema(MediationIdMatchInfoSchema())
@response_schema(MediationRecordSchema(), 200)
async def retrieve_mediation_request(request: web.BaseRequest):
    """Retrieve a single mediation request."""
    context: AdminRequestContext = request["context"]

    mediation_id = request.match_info["mediation_id"]
    try:
        session = await context.session()
        mediation_record = await MediationRecord.retrieve_by_id(session, mediation_id)
        result = mediation_record.serialize()
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except (BaseModelError, StorageError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response(result)


@docs(tags=["mediation"], summary="Delete mediation request by ID")
@match_info_schema(MediationIdMatchInfoSchema())
@response_schema(MediationRecordSchema, 200)
async def delete_mediation_request(request: web.BaseRequest):
    """Delete a mediation request by ID."""
    context: AdminRequestContext = request["context"]

    mediation_id = request.match_info["mediation_id"]
    try:
        session = await context.session()

        mediation_record = await MediationRecord.retrieve_by_id(session, mediation_id)
        result = mediation_record.serialize()
        await mediation_record.delete_record(session)
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except (BaseModelError, StorageError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response(result)


@docs(tags=["mediation"], summary="Request mediation from connection")
@match_info_schema(ConnIdMatchInfoSchema())
@request_schema(MediationCreateRequestSchema())
@response_schema(MediationRecordSchema(), 201)
async def request_mediation(request: web.BaseRequest):
    """Request mediation from connection."""
    context: AdminRequestContext = request["context"]
    outbound_message_router = request["outbound_message_router"]

    conn_id = request.match_info["conn_id"]

    body = await request.json()
    mediator_terms = body.get("mediator_terms")
    recipient_terms = body.get("recipient_terms")

    try:
        async with context.session() as session:
            connection_record = await ConnRecord.retrieve_by_id(session, conn_id)

            if not connection_record.is_ready:
                raise web.HTTPBadRequest(reason="requested connection is not ready")

            if await MediationRecord.exists_for_connection_id(session, conn_id):
                raise web.HTTPBadRequest(
                    reason=f"MediationRecord already exists for connection {conn_id}"
                )

        mediation_record, mediation_request = await MediationManager(
            context.profile
        ).prepare_request(
            connection_id=conn_id,
            mediator_terms=mediator_terms,
            recipient_terms=recipient_terms,
        )

        result = mediation_record.serialize()
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except (StorageError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    await outbound_message_router(mediation_request, connection_id=conn_id)

    return web.json_response(result, status=201)


@docs(tags=["mediation"], summary="Grant received mediation")
@match_info_schema(MediationIdMatchInfoSchema())
@response_schema(MediationGrantSchema(), 201)
async def mediation_request_grant(request: web.BaseRequest):
    """Grant a stored mediation request."""
    context: AdminRequestContext = request["context"]
    outbound_handler = request["outbound_message_router"]
    mediation_id = request.match_info.get("mediation_id")
    try:
        mediation_mgr = MediationManager(context.profile)
        record, grant_request = await mediation_mgr.grant_request(
            mediation_id=mediation_id
        )
        result = record.serialize()
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except (MediationManagerError, StorageError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err
    await outbound_handler(grant_request, connection_id=record.connection_id)
    return web.json_response(result, status=201)


@docs(tags=["mediation"], summary="Deny a stored mediation request")
@match_info_schema(MediationIdMatchInfoSchema())
@request_schema(AdminMediationDenySchema())
@response_schema(MediationDenySchema(), 201)
async def mediation_request_deny(request: web.BaseRequest):
    """Deny a stored mediation request."""
    context: AdminRequestContext = request["context"]
    outbound_handler = request["outbound_message_router"]
    mediation_id = request.match_info.get("mediation_id")
    body = await request.json()
    mediator_terms = body.get("mediator_terms")
    recipient_terms = body.get("recipient_terms")
    try:
        mediation_manager = MediationManager(context.profile)
        record, deny_request = await mediation_manager.deny_request(
            mediation_id=mediation_id,
            mediator_terms=mediator_terms,
            recipient_terms=recipient_terms,
        )
        result = record.serialize()
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except (MediationManagerError, StorageError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    await outbound_handler(deny_request, connection_id=record.connection_id)
    return web.json_response(result, status=201)


@docs(
    tags=["mediation"],
    summary="Retrieve keylists by connection or role",
)
@querystring_schema(GetKeylistQuerySchema())
@response_schema(KeylistSchema(), 200)
async def get_keylist(request: web.BaseRequest):
    """Retrieve keylists by connection or role."""
    context: AdminRequestContext = request["context"]
    connection_id = request.query.get("conn_id")
    role = request.query.get("role")

    tag_filter = {}
    if connection_id:
        tag_filter["connection_id"] = connection_id
    if role:
        tag_filter["role"] = role

    try:
        async with context.session() as session:
            keylists = await RouteRecord.query(session, tag_filter)
        results = [record.serialize() for record in keylists]
    except (StorageError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err
    return web.json_response(results, status=200)


@docs(
    tags=["mediation"],
    summary="Send keylist query to mediator",
)
@match_info_schema(MediationIdMatchInfoSchema())
@querystring_schema(KeylistQueryPaginateQuerySchema())
@request_schema(KeylistQueryFilterRequestSchema())
@response_schema(KeylistQuerySchema(), 201)
async def send_keylist_query(request: web.BaseRequest):
    """Send keylist query to mediator."""
    context: AdminRequestContext = request["context"]
    outbound_handler = request["outbound_message_router"]

    mediation_id = request.match_info["mediation_id"]

    body = await request.json()
    filter_ = body.get("filter")

    paginate_limit = request.query.get("paginate_limit")
    paginate_offset = request.query.get("paginate_offset")

    try:
        async with context.session() as session:
            record = await MediationRecord.retrieve_by_id(session, mediation_id)
        mediation_manager = MediationManager(context.profile)
        request = await mediation_manager.prepare_keylist_query(
            filter_=filter_,
            paginate_limit=paginate_limit,
            paginate_offset=paginate_offset,
        )
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except (StorageError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    await outbound_handler(request, connection_id=record.connection_id)
    return web.json_response(request.serialize(), status=201)


@docs(tags=["mediation"], summary="Send keylist update to mediator")
@match_info_schema(MediationIdMatchInfoSchema())
@request_schema(KeylistUpdateRequestSchema())
@response_schema(KeylistUpdateSchema(), 201)
async def send_keylist_update(request: web.BaseRequest):
    """Send keylist update to mediator."""
    context: AdminRequestContext = request["context"]

    outbound_handler = request["outbound_message_router"]

    mediation_id = request.match_info["mediation_id"]

    body = await request.json()
    updates = body.get("updates")

    if not updates:
        raise web.HTTPBadRequest(reason="Updates cannot be empty.")

    mediation_mgr = MediationManager(context.profile)
    keylist_updates = None
    for update in updates:
        if update.get("action") == KeylistUpdateRule.RULE_ADD:
            keylist_updates = await mediation_mgr.add_key(
                update.get("recipient_key"), keylist_updates
            )
        elif update.get("action") == KeylistUpdateRule.RULE_REMOVE:
            keylist_updates = await mediation_mgr.remove_key(
                update.get("recipient_key"), keylist_updates
            )
        else:
            raise web.HTTPBadRequest(reason="Invalid action for keylist update.")

    try:
        async with context.session() as session:
            record = await MediationRecord.retrieve_by_id(session, mediation_id)
        if record.state != MediationRecord.STATE_GRANTED:
            raise web.HTTPBadRequest(reason=("mediation is not granted."))
        results = keylist_updates.serialize()
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except (StorageError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err
    await outbound_handler(keylist_updates, connection_id=record.connection_id)
    return web.json_response(results, status=201)


@docs(tags=["mediation"], summary="Get default mediator")
@response_schema(MediationRecordSchema(), 200)
async def get_default_mediator(request: web.BaseRequest):
    """Get default mediator."""
    context: AdminRequestContext = request["context"]
    try:
        default_mediator = await MediationManager(
            context.profile
        ).get_default_mediator()
        results = default_mediator.serialize() if default_mediator else {}
    except (StorageError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err
    return web.json_response(results, status=200)


@docs(tags=["mediation"], summary="Set default mediator")
@match_info_schema(MediationIdMatchInfoSchema())
@response_schema(MediationRecordSchema(), 201)
async def set_default_mediator(request: web.BaseRequest):
    """Set default mediator."""
    context: AdminRequestContext = request["context"]
    mediation_id = request.match_info.get("mediation_id")
    try:
        mediator_mgr = MediationManager(context.profile)
        await mediator_mgr.set_default_mediator_by_id(mediation_id=mediation_id)
        default_mediator = await mediator_mgr.get_default_mediator()
        results = default_mediator.serialize()
    except (StorageError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err
    return web.json_response(results, status=201)


@docs(tags=["mediation"], summary="Clear default mediator")
@response_schema(MediationRecordSchema(), 201)
async def clear_default_mediator(request: web.BaseRequest):
    """Clear set default mediator."""
    context: AdminRequestContext = request["context"]
    try:
        mediator_mgr = MediationManager(context.profile)
        default_mediator = await mediator_mgr.get_default_mediator()
        await mediator_mgr.clear_default_mediator()
        results = default_mediator.serialize()
    except (StorageError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err
    return web.json_response(results, status=201)


async def register(app: web.Application):
    """Register routes."""

    app.add_routes(
        [
            web.get("/mediation/requests", list_mediation_requests, allow_head=False),
            web.get(
                "/mediation/requests/{mediation_id}",
                retrieve_mediation_request,
                allow_head=False,
            ),
            web.delete("/mediation/requests/{mediation_id}", delete_mediation_request),
            web.post(
                "/mediation/requests/{mediation_id}/grant",
                mediation_request_grant,
            ),
            web.post("/mediation/requests/{mediation_id}/deny", mediation_request_deny),
            web.post("/mediation/request/{conn_id}", request_mediation),
            web.get("/mediation/keylists", get_keylist, allow_head=False),
            web.post(
                "/mediation/keylists/{mediation_id}/send-keylist-update",
                send_keylist_update,
            ),
            web.post(
                "/mediation/keylists/{mediation_id}/send-keylist-query",
                send_keylist_query,
            ),
            web.get(
                "/mediation/default-mediator", get_default_mediator, allow_head=False
            ),
            web.put("/mediation/{mediation_id}/default-mediator", set_default_mediator),
            web.delete("/mediation/default-mediator", clear_default_mediator),
        ]
    )


def post_process_routes(app: web.Application):
    """Amend swagger API."""

    # Add top-level tags description
    if "tags" not in app._state["swagger_dict"]:
        app._state["swagger_dict"]["tags"] = []
    app._state["swagger_dict"]["tags"].append(
        {
            "name": "mediation",
            "description": "Mediation management",
            "externalDocs": {"description": "Specification", "url": SPEC_URI},
        }
    )



"""DIF-specific credential exchange information with non-secrets storage."""

from typing import Any

from marshmallow import EXCLUDE, fields

from ......core.profile import ProfileSession
from ......messaging.models.base_record import BaseRecord, BaseRecordSchema
from ......messaging.valid import UUIDFour

from .. import UNENCRYPTED_TAGS


class V20CredExRecordDIF(BaseRecord):
    """Credential exchange DIF detail record."""

    class Meta:
        """V20CredExRecordDIF metadata."""

        schema_class = "V20CredExRecordDIFSchema"

    RECORD_ID_NAME = "cred_ex_dif_id"
    RECORD_TYPE = "dif_cred_ex_v20"
    TAG_NAMES = {"~cred_ex_id"} if UNENCRYPTED_TAGS else {"cred_ex_id"}
    WEBHOOK_TOPIC = "issue_credential_v2_0_dif"

    def __init__(
        self,
        cred_ex_dif_id: str = None,
        *,
        cred_ex_id: str = None,
        # TODO: REMOVE THIS COMMENT AND SET DIF ITEMS BELOW
        item: str = None,
        **kwargs,
    ):
        """Initialize DIF credential exchange record details."""
        super().__init__(cred_ex_dif_id, **kwargs)

        self.cred_ex_id = cred_ex_id
        # TODO: REMOVE THIS COMMENT AND SET DIF ITEMS BELOW
        self.item = item

    @property
    def cred_ex_dif_id(self) -> str:
        """Accessor for the ID associated with this exchange."""
        return self._id

    @property
    def record_value(self) -> dict:
        """Accessor for the JSON record value generated for this credential exchange."""
        return {
            prop: getattr(self, prop)
            for prop in (
                # TODO: REMOVE THIS COMMENT AND SET DIF ITEMS BELOW
                "item",
            )
        }

    @classmethod
    async def retrieve_by_cred_ex_id(
        cls,
        session: ProfileSession,
        cred_ex_id: str,
    ) -> "V20CredExRecordDIF":
        """Retrieve a credential exchange DIF detail record by its cred ex id."""
        return await cls.retrieve_by_tag_filter(
            session,
            {"cred_ex_id": cred_ex_id},
            None,
        )

    def __eq__(self, other: Any) -> bool:
        """Comparison between records."""
        return super().__eq__(other)


class V20CredExRecordDIFSchema(BaseRecordSchema):
    """Credential exchange DIF detail record detail schema."""

    class Meta:
        """Credential exchange DIF detail record schema metadata."""

        model_class = V20CredExRecordDIF
        unknown = EXCLUDE

    cred_ex_dif_id = fields.Str(
        required=False,
        description="Record identifier",
        example=UUIDFour.EXAMPLE,
    )
    cred_ex_id = fields.Str(
        required=False,
        description="Corresponding v2.0 credential exchange record identifier",
        example=UUIDFour.EXAMPLE,
    )
    # TODO: REMOVE THIS COMMENT AND SET DIF ITEMS BELOW
    item = fields.Dict(
        required=False,
        description="DIF item",
        example=UUIDFour.EXAMPLE,
    )



import pytest
from asynctest import (
    mock as async_mock,
    TestCase as AsyncTestCase,
)

from ......messaging.request_context import RequestContext
from ......messaging.responder import MockResponder

from .. import menu_handler as handler


class TestMenuHandler(AsyncTestCase):
    async def test_called(self):
        request_context = RequestContext.test_context()
        request_context.connection_record = async_mock.MagicMock()
        request_context.connection_record.connection_id = "dummy"

        handler.save_connection_menu = async_mock.CoroutineMock()
        responder = MockResponder()

        request_context.message = handler.Menu()
        handler_inst = handler.MenuHandler()
        await handler_inst.handle(request_context, responder)

        handler.save_connection_menu.assert_called_once_with(
            request_context.message,
            request_context.connection_record.connection_id,
            request_context,
        )



"""Standard packed message format classes."""

from base64 import b64decode
import json
import logging
from typing import List, Sequence, Tuple, Union

from ..core.profile import ProfileSession

from ..protocols.routing.v1_0.messages.forward import Forward

from ..messaging.util import time_now
from ..utils.task_queue import TaskQueue
from ..wallet.base import BaseWallet
from ..wallet.error import WalletError

from .error import WireFormatParseError, WireFormatEncodeError, RecipientKeysError
from .inbound.receipt import MessageReceipt
from .wire_format import BaseWireFormat

LOGGER = logging.getLogger(__name__)


class PackWireFormat(BaseWireFormat):
    """Standard DIDComm message parser and serializer."""

    def __init__(self):
        """Initialize the pack wire format instance."""
        super().__init__()
        self.task_queue: TaskQueue = None

    async def parse_message(
        self,
        session: ProfileSession,
        message_body: Union[str, bytes],
    ) -> Tuple[dict, MessageReceipt]:
        """
        Deserialize an incoming message and further populate the request context.

        Args:
            session: The profile session for providing wallet access
            message_body: The body of the message

        Returns:
            A tuple of the parsed message and a message receipt instance

        Raises:
            WireFormatParseError: If the JSON parsing failed
            WireFormatParseError: If a wallet is required but can't be located

        """

        receipt = MessageReceipt()
        receipt.in_time = time_now()
        receipt.raw_message = message_body

        message_dict = None
        message_json = message_body

        if not message_json:
            raise WireFormatParseError("Message body is empty")

        try:
            message_dict = json.loads(message_json)
        except ValueError:
            raise WireFormatParseError("Message JSON parsing failed")
        if not isinstance(message_dict, dict):
            raise WireFormatParseError("Message JSON result is not an object")

        # packed messages are detected by the absence of @type
        if "@type" not in message_dict:

            try:
                unpack = self.unpack(session, message_body, receipt)
                message_json = await (
                    self.task_queue and self.task_queue.run(unpack) or unpack
                )
            except WireFormatParseError:
                LOGGER.debug("Message unpack failed, falling back to JSON")
            else:
                receipt.raw_message = message_json
                try:
                    message_dict = json.loads(message_json)
                except ValueError:
                    raise WireFormatParseError("Message JSON parsing failed")
                if not isinstance(message_dict, dict):
                    raise WireFormatParseError("Message JSON result is not an object")

        # parse thread ID
        thread_dec = message_dict.get("~thread")
        receipt.thread_id = (
            thread_dec and thread_dec.get("thid") or message_dict.get("@id")
        )

        # handle transport decorator
        transport_dec = message_dict.get("~transport")
        if transport_dec:
            receipt.direct_response_mode = transport_dec.get("return_route")

        LOGGER.debug(f"Expanded message: {message_dict}")

        return message_dict, receipt

    async def unpack(
        self,
        session: ProfileSession,
        message_body: Union[str, bytes],
        receipt: MessageReceipt,
    ):
        """Look up the wallet instance and perform the message unpack."""
        wallet = session.inject(BaseWallet, required=False)
        if not wallet:
            raise WireFormatParseError("Wallet not defined in profile session")

        try:
            unpacked = await wallet.unpack_message(message_body)
            (
                message_json,
                receipt.sender_verkey,
                receipt.recipient_verkey,
            ) = unpacked
            return message_json
        except WalletError as e:
            raise WireFormatParseError("Message unpack failed") from e

    async def encode_message(
        self,
        session: ProfileSession,
        message_json: Union[str, bytes],
        recipient_keys: Sequence[str],
        routing_keys: Sequence[str],
        sender_key: str,
    ) -> Union[str, bytes]:
        """
        Encode an outgoing message for transport.

        Args:
            session: The profile session for providing wallet access
            message_json: The message body to serialize
            recipient_keys: A sequence of recipient verkeys
            routing_keys: A sequence of routing verkeys
            sender_key: The verification key of the sending agent

        Returns:
            The encoded message

        Raises:
            MessageEncodeError: If the message could not be encoded

        """

        if sender_key and recipient_keys:
            pack = self.pack(
                session, message_json, recipient_keys, routing_keys, sender_key
            )
            message = await (self.task_queue and self.task_queue.run(pack) or pack)
        else:
            message = message_json
        return message

    async def pack(
        self,
        session: ProfileSession,
        message_json: Union[str, bytes],
        recipient_keys: Sequence[str],
        routing_keys: Sequence[str],
        sender_key: str,
    ):
        """Look up the wallet instance and perform the message pack."""
        if not sender_key or not recipient_keys:
            raise WireFormatEncodeError("Cannot pack message without associated keys")

        wallet = session.inject(BaseWallet, required=False)
        if not wallet:
            raise WireFormatEncodeError("No wallet instance")

        try:
            message = await wallet.pack_message(
                message_json, recipient_keys, sender_key
            )
        except WalletError as e:
            raise WireFormatEncodeError("Message pack failed") from e

        if routing_keys:
            recip_keys = recipient_keys
            for router_key in routing_keys:
                message = json.loads(message.decode("utf-8"))
                fwd_msg = Forward(to=recip_keys[0], msg=message)
                # Forwards are anon packed
                recip_keys = [router_key]
                try:
                    message = await wallet.pack_message(fwd_msg.to_json(), recip_keys)
                except WalletError as e:
                    raise WireFormatEncodeError("Forward message pack failed") from e
        return message

    def get_recipient_keys(self, message_body: Union[str, bytes]) -> List[str]:
        """
        Get all recipient keys from a wire message.

        Args:
            message_body: The body of the message

        Returns:
            List of recipient keys from the message body

        Raises:
            RecipientKeysError: If the recipient keys could not be extracted

        """

        try:
            message_dict = json.loads(message_body)
            protected = json.loads(b64decode(message_dict["protected"]))
            recipients = protected["recipients"]

            recipient_keys = [recipient["header"]["kid"] for recipient in recipients]
        except Exception as e:
            raise RecipientKeysError(
                "Error trying to extract recipient keys from JWE", e
            )

        return recipient_keys



"""Indy SDK holder implementation."""

import json
import logging

from collections import OrderedDict
from typing import Sequence, Tuple, Union

import indy.anoncreds
from indy.error import ErrorCode, IndyError

from ...indy.sdk.wallet_setup import IndyOpenWallet
from ...ledger.base import BaseLedger
from ...storage.indy import IndySdkStorage
from ...storage.error import StorageError, StorageNotFoundError
from ...storage.record import StorageRecord
from ...wallet.error import WalletNotFoundError

from ..holder import IndyHolder, IndyHolderError

from .error import IndyErrorHandler
from .util import create_tails_reader

LOGGER = logging.getLogger(__name__)


class IndySdkHolder(IndyHolder):
    """Indy-SDK holder implementation."""

    def __init__(self, wallet: IndyOpenWallet):
        """
        Initialize an IndyHolder instance.

        Args:
            wallet: IndyOpenWallet instance

        """
        self.wallet = wallet

    async def create_credential_request(
        self, credential_offer: dict, credential_definition: dict, holder_did: str
    ) -> Tuple[str, str]:
        """
        Create a credential request for the given credential offer.

        Args:
            credential_offer: The credential offer to create request for
            credential_definition: The credential definition to create an offer for
            holder_did: the DID of the agent making the request

        Returns:
            A tuple of the credential request and credential request metadata

        """

        with IndyErrorHandler(
            "Error when creating credential request", IndyHolderError
        ):
            (
                credential_request_json,
                credential_request_metadata_json,
            ) = await indy.anoncreds.prover_create_credential_req(
                self.wallet.handle,
                holder_did,
                json.dumps(credential_offer),
                json.dumps(credential_definition),
                self.wallet.master_secret_id,
            )

        LOGGER.debug(
            "Created credential request. "
            "credential_request_json=%s credential_request_metadata_json=%s",
            credential_request_json,
            credential_request_metadata_json,
        )

        return credential_request_json, credential_request_metadata_json

    async def store_credential(
        self,
        credential_definition: dict,
        credential_data: dict,
        credential_request_metadata: dict,
        credential_attr_mime_types=None,
        credential_id: str = None,
        rev_reg_def: dict = None,
    ) -> str:
        """
        Store a credential in the wallet.

        Args:
            credential_definition: Credential definition for this credential
            credential_data: Credential data generated by the issuer
            credential_request_metadata: credential request metadata generated
                by the issuer
            credential_attr_mime_types: dict mapping attribute names to (optional)
                MIME types to store as non-secret record, if specified
            credential_id: optionally override the stored credential id
            rev_reg_def: revocation registry definition in json

        Returns:
            the ID of the stored credential

        """
        with IndyErrorHandler(
            "Error when storing credential in wallet", IndyHolderError
        ):
            credential_id = await indy.anoncreds.prover_store_credential(
                wallet_handle=self.wallet.handle,
                cred_id=credential_id,
                cred_req_metadata_json=json.dumps(credential_request_metadata),
                cred_json=json.dumps(credential_data),
                cred_def_json=json.dumps(credential_definition),
                rev_reg_def_json=json.dumps(rev_reg_def) if rev_reg_def else None,
            )

        if credential_attr_mime_types:
            mime_types = {
                attr: credential_attr_mime_types.get(attr)
                for attr in credential_data["values"]
                if attr in credential_attr_mime_types
            }
            if mime_types:
                record = StorageRecord(
                    type=IndyHolder.RECORD_TYPE_MIME_TYPES,
                    value=credential_id,
                    tags=mime_types,
                    id=f"{IndyHolder.RECORD_TYPE_MIME_TYPES}::{credential_id}",
                )
                indy_stor = IndySdkStorage(self.wallet)
                await indy_stor.add_record(record)

        return credential_id

    async def get_credentials(self, start: int, count: int, wql: dict):
        """
        Get credentials stored in the wallet.

        Args:
            start: Starting index
            count: Number of records to return
            wql: wql query dict

        """

        async def fetch(limit):
            """Fetch up to limit (default smaller of all remaining or 256) creds."""
            creds = []
            CHUNK = min(record_count, limit or record_count, IndyHolder.CHUNK)
            cardinality = min(limit or record_count, record_count)

            with IndyErrorHandler(
                "Error fetching credentials from wallet", IndyHolderError
            ):
                while len(creds) < cardinality:
                    batch = json.loads(
                        await indy.anoncreds.prover_fetch_credentials(
                            search_handle, CHUNK
                        )
                    )
                    creds.extend(batch)
                    if len(batch) < CHUNK:
                        break
            return creds

        with IndyErrorHandler(
            "Error when constructing wallet credential query", IndyHolderError
        ):
            (
                search_handle,
                record_count,
            ) = await indy.anoncreds.prover_search_credentials(
                self.wallet.handle, json.dumps(wql)
            )

            if start > 0:
                # must move database cursor manually
                await fetch(start)
            credentials = await fetch(count)

            await indy.anoncreds.prover_close_credentials_search(search_handle)

        return credentials

    async def get_credentials_for_presentation_request_by_referent(
        self,
        presentation_request: dict,
        referents: Sequence[str],
        start: int,
        count: int,
        extra_query: dict = {},
    ):
        """
        Get credentials stored in the wallet.

        Args:
            presentation_request: Valid presentation request from issuer
            referents: Presentation request referents to use to search for creds
            start: Starting index
            count: Maximum number of records to return
            extra_query: wql query dict

        """

        async def fetch(reft, limit):
            """Fetch up to limit (default smaller of all remaining or 256) creds."""
            creds = []
            CHUNK = min(IndyHolder.CHUNK, limit or IndyHolder.CHUNK)

            with IndyErrorHandler(
                "Error fetching credentials from wallet for presentation request",
                IndyHolderError,
            ):
                while not limit or len(creds) < limit:
                    batch = json.loads(
                        await indy.anoncreds.prover_fetch_credentials_for_proof_req(
                            search_handle, reft, CHUNK
                        )
                    )
                    creds.extend(batch)
                    if len(batch) < CHUNK:
                        break
            return creds

        with IndyErrorHandler(
            "Error when constructing wallet credential query", IndyHolderError
        ):
            search_handle = await (
                indy.anoncreds.prover_search_credentials_for_proof_req(
                    self.wallet.handle,
                    json.dumps(presentation_request),
                    json.dumps(extra_query),
                )
            )

            if not referents:
                referents = (
                    *presentation_request["requested_attributes"],
                    *presentation_request["requested_predicates"],
                )
            creds_dict = OrderedDict()

            try:
                for reft in referents:
                    # must move database cursor manually
                    if start > 0:
                        await fetch(reft, start)
                    credentials = await fetch(reft, count - len(creds_dict))

                    for cred in credentials:
                        cred_id = cred["cred_info"]["referent"]
                        if cred_id not in creds_dict:
                            cred["presentation_referents"] = {reft}
                            creds_dict[cred_id] = cred
                        else:
                            creds_dict[cred_id]["presentation_referents"].add(reft)
                    if len(creds_dict) >= count:
                        break
            finally:
                # Always close
                await indy.anoncreds.prover_close_credentials_search_for_proof_req(
                    search_handle
                )

        for cred in creds_dict.values():
            cred["presentation_referents"] = list(cred["presentation_referents"])

        creds_ordered = tuple(
            [
                cred
                for cred in sorted(
                    creds_dict.values(),
                    key=lambda c: (
                        c["cred_info"]["rev_reg_id"] or "",  # irrevocable 1st
                        c["cred_info"][
                            "referent"
                        ],  # should be descending by timestamp if we had it
                    ),
                )
            ]
        )[:count]
        return creds_ordered

    async def get_credential(self, credential_id: str) -> str:
        """
        Get a credential stored in the wallet.

        Args:
            credential_id: Credential id to retrieve

        """
        try:
            credential_json = await indy.anoncreds.prover_get_credential(
                self.wallet.handle, credential_id
            )
        except IndyError as err:
            if err.error_code == ErrorCode.WalletItemNotFound:
                raise WalletNotFoundError(
                    "Credential {} not found in wallet {}".format(
                        credential_id, self.wallet.name
                    )
                )
            else:
                raise IndyErrorHandler.wrap_error(
                    err,
                    f"Error when fetching credential {credential_id}",
                    IndyHolderError,
                ) from err

        return credential_json

    async def credential_revoked(
        self, ledger: BaseLedger, credential_id: str, fro: int = None, to: int = None
    ) -> bool:
        """
        Check ledger for revocation status of credential by cred id.

        Args:
            credential_id: Credential id to check

        """
        cred = json.loads(await self.get_credential(credential_id))
        rev_reg_id = cred["rev_reg_id"]

        if rev_reg_id:
            cred_rev_id = int(cred["cred_rev_id"])
            (rev_reg_delta, _) = await ledger.get_revoc_reg_delta(
                rev_reg_id,
                fro,
                to,
            )

            return cred_rev_id in rev_reg_delta["value"].get("revoked", [])
        else:
            return False

    async def delete_credential(self, credential_id: str):
        """
        Remove a credential stored in the wallet.

        Args:
            credential_id: Credential id to remove

        """
        try:
            indy_stor = IndySdkStorage(self.wallet)
            mime_types_record = await indy_stor.get_record(
                IndyHolder.RECORD_TYPE_MIME_TYPES,
                f"{IndyHolder.RECORD_TYPE_MIME_TYPES}::{credential_id}",
            )
            await indy_stor.delete_record(mime_types_record)
        except StorageNotFoundError:
            pass  # MIME types record not present: carry on

        try:
            await indy.anoncreds.prover_delete_credential(
                self.wallet.handle, credential_id
            )
        except IndyError as err:
            if err.error_code == ErrorCode.WalletItemNotFound:
                raise WalletNotFoundError(
                    "Credential {} not found in wallet {}".format(
                        credential_id, self.wallet.name
                    )
                )
            else:
                raise IndyErrorHandler.wrap_error(
                    err, "Error when deleting credential", IndyHolderError
                ) from err

    async def get_mime_type(
        self, credential_id: str, attr: str = None
    ) -> Union[dict, str]:
        """
        Get MIME type per attribute (or for all attributes).

        Args:
            credential_id: credential id
            attr: attribute of interest or omit for all

        Returns: Attribute MIME type or dict mapping attribute names to MIME types
            attr_meta_json = all_meta.tags.get(attr)

        """
        try:
            mime_types_record = await IndySdkStorage(self.wallet).get_record(
                IndyHolder.RECORD_TYPE_MIME_TYPES,
                f"{IndyHolder.RECORD_TYPE_MIME_TYPES}::{credential_id}",
            )
        except StorageError:
            return None  # no MIME types: not an error

        return mime_types_record.tags.get(attr) if attr else mime_types_record.tags

    async def create_presentation(
        self,
        presentation_request: dict,
        requested_credentials: dict,
        schemas: dict,
        credential_definitions: dict,
        rev_states: dict = None,
    ) -> str:
        """
        Get credentials stored in the wallet.

        Args:
            presentation_request: Valid indy format presentation request
            requested_credentials: Indy format requested credentials
            schemas: Indy formatted schemas JSON
            credential_definitions: Indy formatted credential definitions JSON
            rev_states: Indy format revocation states JSON

        """

        with IndyErrorHandler("Error when constructing proof", IndyHolderError):
            presentation_json = await indy.anoncreds.prover_create_proof(
                self.wallet.handle,
                json.dumps(presentation_request),
                json.dumps(requested_credentials),
                self.wallet.master_secret_id,
                json.dumps(schemas),
                json.dumps(credential_definitions),
                json.dumps(rev_states) if rev_states else "{}",
            )

        return presentation_json

    async def create_revocation_state(
        self,
        cred_rev_id: str,
        rev_reg_def: dict,
        rev_reg_delta: dict,
        timestamp: int,
        tails_file_path: str,
    ) -> str:
        """
        Create current revocation state for a received credential.

        Args:
            cred_rev_id: credential revocation id in revocation registry
            rev_reg_def: revocation registry definition
            rev_reg_delta: revocation delta
            timestamp: delta timestamp

        Returns:
            the revocation state

        """

        with IndyErrorHandler(
            "Error when constructing revocation state", IndyHolderError
        ):
            tails_file_reader = await create_tails_reader(tails_file_path)
            rev_state_json = await indy.anoncreds.create_revocation_state(
                tails_file_reader,
                rev_reg_def_json=json.dumps(rev_reg_def),
                cred_rev_id=cred_rev_id,
                rev_reg_delta_json=json.dumps(rev_reg_delta),
                timestamp=timestamp,
            )

        return rev_state_json



import pytest
from asynctest import mock as async_mock

from ......messaging.request_context import RequestContext
from ......messaging.responder import MockResponder
from ......transport.inbound.receipt import MessageReceipt

from ...manager import DIDXManagerError
from ...messages.complete import DIDXComplete
from ...messages.problem_report import ProblemReportReason

from .. import complete_handler as test_module


@pytest.fixture()
def request_context() -> RequestContext:
    ctx = RequestContext.test_context()
    ctx.message_receipt = MessageReceipt()
    yield ctx


class TestDIDXCompleteHandler:
    """Class unit testing complete handler."""

    @pytest.mark.asyncio
    @async_mock.patch.object(test_module, "DIDXManager")
    async def test_called(self, mock_conn_mgr, request_context):
        mock_conn_mgr.return_value.accept_complete = async_mock.CoroutineMock()
        request_context.message = DIDXComplete()
        handler_inst = test_module.DIDXCompleteHandler()
        responder = MockResponder()
        await handler_inst.handle(request_context, None)

        mock_conn_mgr.return_value.accept_complete.assert_called_once_with(
            request_context.message, request_context.message_receipt
        )

    @pytest.mark.asyncio
    @async_mock.patch.object(test_module, "DIDXManager")
    async def test_x(self, mock_conn_mgr, request_context):
        mock_conn_mgr.return_value.accept_complete = async_mock.CoroutineMock(
            side_effect=DIDXManagerError(
                error_code=ProblemReportReason.COMPLETE_NOT_ACCEPTED
            )
        )
        mock_conn_mgr.return_value._logger = async_mock.MagicMock(
            exception=async_mock.MagicMock()
        )
        request_context.message = DIDXComplete()
        handler_inst = test_module.DIDXCompleteHandler()
        responder = MockResponder()
        await handler_inst.handle(request_context, responder)

        assert mock_conn_mgr.return_value._logger.exception.called_once_(
            "Error receiving connection complete"
        )



import asyncio
import pytest

from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop
from aiohttp import web
from asynctest import mock as async_mock

from ....core.in_memory import InMemoryProfile
from ....utils.stats import Collector

from ...wire_format import JsonWireFormat

from ..base import OutboundTransportError
from ..http import HttpTransport


class TestHttpTransport(AioHTTPTestCase):
    async def setUpAsync(self):
        self.profile = InMemoryProfile.test_profile()
        self.message_results = []
        self.headers = {}

    async def receive_message(self, request):
        payload = await request.json()
        self.headers = request.headers
        self.message_results.append(payload)
        raise web.HTTPOk()

    async def get_application(self):
        """
        Override the get_app method to return your application.
        """
        app = web.Application()
        app.add_routes([web.post("/", self.receive_message)])
        return app

    @unittest_run_loop
    async def test_handle_message_no_api_key(self):
        server_addr = f"http://localhost:{self.server.port}"

        async def send_message(transport, payload, endpoint):
            async with transport:
                await transport.handle_message(self.profile, payload, endpoint)

        transport = HttpTransport()

        await asyncio.wait_for(send_message(transport, "{}", endpoint=server_addr), 5.0)
        assert self.message_results == [{}]
        assert self.headers.get("x-api-key") is None
        assert self.headers.get("content-type") == "application/json"

    @unittest_run_loop
    async def test_handle_message_api_key(self):
        server_addr = f"http://localhost:{self.server.port}"
        api_key = "test1234"

        async def send_message(transport, payload, endpoint, api_key):
            async with transport:
                await transport.handle_message(
                    self.profile, payload, endpoint, api_key=api_key
                )

        transport = HttpTransport()

        await asyncio.wait_for(
            send_message(transport, "{}", endpoint=server_addr, api_key=api_key), 5.0
        )
        assert self.message_results == [{}]
        assert self.headers.get("x-api-key") == api_key

    @unittest_run_loop
    async def test_handle_message_packed_compat_mime_type(self):
        server_addr = f"http://localhost:{self.server.port}"

        async def send_message(transport, payload, endpoint):
            async with transport:
                await transport.handle_message(self.profile, payload, endpoint)

        transport = HttpTransport()

        await asyncio.wait_for(
            send_message(transport, b"{}", endpoint=server_addr), 5.0
        )
        assert self.message_results == [{}]
        assert self.headers.get("content-type") == "application/ssi-agent-wire"

    @unittest_run_loop
    async def test_handle_message_packed_standard_mime_type(self):
        server_addr = f"http://localhost:{self.server.port}"

        async def send_message(transport, payload, endpoint):
            async with transport:
                await transport.handle_message(self.profile, payload, endpoint)

        transport = HttpTransport()

        self.profile.settings["emit_new_didcomm_mime_type"] = True
        await asyncio.wait_for(
            send_message(transport, b"{}", endpoint=server_addr), 5.0
        )
        assert self.message_results == [{}]
        assert self.headers.get("content-type") == "application/didcomm-envelope-enc"

    @unittest_run_loop
    async def test_stats(self):
        server_addr = f"http://localhost:{self.server.port}"

        async def send_message(transport, payload, endpoint):
            async with transport:
                await transport.handle_message(self.profile, payload, endpoint)

        transport = HttpTransport()
        transport.collector = Collector()
        await asyncio.wait_for(
            send_message(transport, b"{}", endpoint=server_addr), 5.0
        )

        results = transport.collector.extract()
        assert results["count"] == {
            "outbound-http:dns_resolve": 1,
            "outbound-http:connect": 1,
            "outbound-http:POST": 1,
        }

    @unittest_run_loop
    async def test_transport_coverage(self):
        transport = HttpTransport()
        assert transport.wire_format is None
        transport.wire_format = JsonWireFormat()
        assert transport.wire_format is not None

        await transport.start()

        with pytest.raises(OutboundTransportError):
            await transport.handle_message(None, None, None)

        with async_mock.patch.object(
            transport, "client_session", async_mock.MagicMock()
        ) as mock_session:
            mock_response = async_mock.MagicMock(status=404)
            mock_session.post = async_mock.MagicMock(
                return_value=async_mock.MagicMock(
                    __aenter__=async_mock.CoroutineMock(return_value=mock_response)
                )
            )
            with pytest.raises(OutboundTransportError):
                await transport.handle_message(None, "dummy", "http://localhost")

        await transport.__aexit__(KeyError, KeyError("just a drill"), None)



"""Credential offer message."""

from typing import Sequence

from marshmallow import EXCLUDE, fields

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from .....messaging.decorators.attach_decorator import (
    AttachDecorator,
    AttachDecoratorSchema,
)
from .....messaging.valid import UUIDFour

from ..message_types import CRED_20_OFFER, PROTOCOL_PACKAGE

from .cred_format import V20CredFormat, V20CredFormatSchema
from .inner.cred_preview import V20CredPreview, V20CredPreviewSchema

HANDLER_CLASS = f"{PROTOCOL_PACKAGE}.handlers.cred_offer_handler.V20CredOfferHandler"


class V20CredOffer(AgentMessage):
    """Credential offer."""

    class Meta:
        """V20CredOffer metadata."""

        handler_class = HANDLER_CLASS
        schema_class = "V20CredOfferSchema"
        message_type = CRED_20_OFFER

    def __init__(
        self,
        _id: str = None,
        *,
        replacement_id: str = None,
        comment: str = None,
        credential_preview: V20CredPreview = None,
        formats: Sequence[V20CredFormat] = None,
        offers_attach: Sequence[AttachDecorator] = None,
        **kwargs,
    ):
        """
        Initialize credential offer object.

        Args:
            replacement_id: unique to issuer, to coordinate credential replacement
            comment: optional human-readable comment
            credential_preview: credential preview
            offers_attach: list of offer attachments

        """
        super().__init__(_id=_id, **kwargs)
        self.replacement_id = replacement_id
        self.comment = comment
        self.credential_preview = credential_preview
        self.formats = list(formats) if formats else []
        self.offers_attach = list(offers_attach) if offers_attach else []

    def offer(self, fmt: V20CredFormat.Format = None) -> dict:
        """
        Return attached offer.

        Args:
            fmt: format of attachment in list to decode and return

        """
        return (fmt or V20CredFormat.Format.INDY).get_attachment_data(
            self.formats,
            self.offers_attach,
        )


class V20CredOfferSchema(AgentMessageSchema):
    """Credential offer schema."""

    class Meta:
        """Credential offer schema metadata."""

        model_class = V20CredOffer
        unknown = EXCLUDE

    replacement_id = fields.Str(
        description="Issuer-unique identifier to coordinate credential replacement",
        required=False,
        allow_none=False,
        example=UUIDFour.EXAMPLE,
    )
    comment = fields.Str(
        description="Human-readable comment",
        required=False,
        allow_none=True,
    )
    credential_preview = fields.Nested(V20CredPreviewSchema, required=False)
    formats = fields.Nested(
        V20CredFormatSchema,
        many=True,
        required=True,
        description="Acceptable credential formats",
    )
    offers_attach = fields.Nested(
        AttachDecoratorSchema, required=True, many=True, data_key="offers~attach"
    )



"""Test Problem Report Message."""
import pytest

from asynctest import TestCase as AsyncTestCase
from unittest import mock, TestCase

from ......messaging.models.base import BaseModelError

from ..problem_report import ProblemReport, ProblemReportReason


class TestProblemReportMessage(TestCase):
    """Test request schema."""

    def setUp(self):
        self.problem_report = ProblemReport(
            problem_code=ProblemReportReason.EXISTING_CONNECTION_NOT_ACTIVE.value,
            explain="Test",
        )

    def test_init(self):
        """Test initialization of Handshake Reuse message."""
        self.problem_report.assign_thread_id(thid="test_thid", pthid="test_pthid")
        assert isinstance(self.problem_report, ProblemReport)
        assert isinstance(self.problem_report._id, str)
        assert len(self.problem_report._id) > 4
        assert self.problem_report._thread.thid == "test_thid"
        assert self.problem_report._thread.pthid == "test_pthid"

    def test_make_model(self):
        """Make problem report model."""
        self.problem_report.assign_thread_id(thid="test_thid", pthid="test_pthid")
        data = self.problem_report.serialize()
        model_instance = ProblemReport.deserialize(data)
        assert isinstance(model_instance, ProblemReport)

    def test_pre_dump_x(self):
        """Exercise pre-dump serialization requirements."""
        with pytest.raises(BaseModelError):
            data = self.problem_report.serialize()



"""Wallet configuration."""

import logging

from ..core.error import ProfileNotFoundError
from ..core.profile import Profile, ProfileManager
from ..wallet.base import BaseWallet, DIDInfo
from ..wallet.crypto import seed_to_did

from .base import ConfigError
from .injection_context import InjectionContext

LOGGER = logging.getLogger(__name__)

CFG_MAP = {"key", "rekey", "name", "storage_config", "storage_creds", "storage_type"}


async def wallet_config(
    context: InjectionContext, provision: bool = False
) -> (Profile, DIDInfo):
    """Initialize the root profile."""

    mgr = context.inject(ProfileManager)

    settings = context.settings
    profile_cfg = {}
    for k in CFG_MAP:
        pk = f"wallet.{k}"
        if pk in settings:
            profile_cfg[k] = settings[pk]

    # may be set by `aca-py provision --recreate`
    if settings.get("wallet.recreate"):
        profile_cfg["auto_recreate"] = True

    if provision:
        profile = await mgr.provision(context, profile_cfg)
    else:
        try:
            profile = await mgr.open(context, profile_cfg)
        except ProfileNotFoundError:
            if settings.get("auto_provision", False):
                profile = await mgr.provision(context, profile_cfg)
            else:
                raise

    if provision:
        if profile.created:
            print("Created new profile")
        else:
            print("Opened existing profile")
        print("Profile backend:", profile.backend)
        print("Profile name:", profile.name)

    wallet_seed = context.settings.get("wallet.seed")
    wallet_local_did = context.settings.get("wallet.local_did")
    txn = await profile.transaction()
    wallet = txn.inject(BaseWallet)

    public_did_info = await wallet.get_public_did()
    public_did = None

    if public_did_info:
        public_did = public_did_info.did
        if wallet_seed and seed_to_did(wallet_seed) != public_did:
            if context.settings.get("wallet.replace_public_did"):
                replace_did_info = await wallet.create_local_did(wallet_seed)
                public_did = replace_did_info.did
                await wallet.set_public_did(public_did)
                print(f"Created new public DID: {public_did}")
                print(f"Verkey: {replace_did_info.verkey}")
            else:
                # If we already have a registered public did and it doesn't match
                # the one derived from `wallet_seed` then we error out.
                raise ConfigError(
                    "New seed provided which doesn't match the registered"
                    + f" public did {public_did}"
                )
        # wait until ledger config to set public DID endpoint - wallet goes first
    elif wallet_seed:
        if wallet_local_did:
            endpoint = context.settings.get("default_endpoint")
            metadata = {"endpoint": endpoint} if endpoint else None

            local_did_info = await wallet.create_local_did(
                seed=wallet_seed, metadata=metadata
            )
            local_did = local_did_info.did
            if provision:
                print(f"Created new local DID: {local_did}")
                print(f"Verkey: {local_did_info.verkey}")
        else:
            public_did_info = await wallet.create_public_did(seed=wallet_seed)
            public_did = public_did_info.did
            if provision:
                print(f"Created new public DID: {public_did}")
                print(f"Verkey: {public_did_info.verkey}")
            # wait until ledger config to set public DID endpoint - wallet goes first

    if provision and not wallet_local_did and not public_did:
        print("No public DID")

    # Debug settings
    test_seed = context.settings.get("debug.seed")
    if context.settings.get("debug.enabled"):
        if not test_seed:
            test_seed = "testseed000000000000000000000001"
    if test_seed:
        await wallet.create_local_did(
            seed=test_seed, metadata={"endpoint": "1.2.3.4:8021"}
        )

    await txn.commit()

    return (profile, public_did_info)



"""
Class to provide some common utilities.

For Connection, DIDExchange and OutOfBand Manager.
"""

import logging

from typing import Sequence, Tuple, List

from ..core.error import BaseError
from ..core.profile import ProfileSession
from ..ledger.base import BaseLedger
from ..protocols.connections.v1_0.messages.connection_invitation import (
    ConnectionInvitation,
)
from ..protocols.coordinate_mediation.v1_0.models.mediation_record import (
    MediationRecord,
)
from ..storage.base import BaseStorage
from ..storage.error import StorageNotFoundError
from ..storage.record import StorageRecord
from ..wallet.base import BaseWallet, DIDInfo
from ..wallet.util import did_key_to_naked

from .models.conn_record import ConnRecord
from .models.connection_target import ConnectionTarget
from .models.diddoc import (
    DIDDoc,
    PublicKey,
    PublicKeyType,
    Service,
)


class BaseConnectionManagerError(BaseError):
    """BaseConnectionManager error."""


class BaseConnectionManager:
    """Class to provide utilities regarding connection_targets."""

    RECORD_TYPE_DID_DOC = "did_doc"
    RECORD_TYPE_DID_KEY = "did_key"

    def __init__(self, session: ProfileSession):
        """
        Initialize a BaseConnectionManager.

        Args:
            session: The profile session for this presentation
        """
        self._logger = logging.getLogger(__name__)
        self._session = session

    async def create_did_document(
        self,
        did_info: DIDInfo,
        inbound_connection_id: str = None,
        svc_endpoints: Sequence[str] = None,
        mediation_records: List[MediationRecord] = None,
    ) -> DIDDoc:
        """Create our DID doc for a given DID.

        Args:
            did_info: The DID information (DID and verkey) used in the connection
            inbound_connection_id: The ID of the inbound routing connection to use
            svc_endpoints: Custom endpoints for the DID Document
            mediation_record: The record for mediation that contains routing_keys and
                service endpoint

        Returns:
            The prepared `DIDDoc` instance

        """

        did_doc = DIDDoc(did=did_info.did)
        did_controller = did_info.did
        did_key = did_info.verkey
        pk = PublicKey(
            did_info.did,
            "1",
            did_key,
            PublicKeyType.ED25519_SIG_2018,
            did_controller,
            True,
        )
        did_doc.set(pk)

        router_id = inbound_connection_id
        routing_keys = []
        router_idx = 1
        while router_id:
            # look up routing connection information
            router = await ConnRecord.retrieve_by_id(self._session, router_id)
            if ConnRecord.State.get(router.state) != ConnRecord.State.COMPLETED:
                raise BaseConnectionManagerError(
                    f"Router connection not completed: {router_id}"
                )
            routing_doc, _ = await self.fetch_did_document(router.their_did)
            if not routing_doc.service:
                raise BaseConnectionManagerError(
                    f"No services defined by routing DIDDoc: {router_id}"
                )
            for service in routing_doc.service.values():
                if not service.endpoint:
                    raise BaseConnectionManagerError(
                        "Routing DIDDoc service has no service endpoint"
                    )
                if not service.recip_keys:
                    raise BaseConnectionManagerError(
                        "Routing DIDDoc service has no recipient key(s)"
                    )
                rk = PublicKey(
                    did_info.did,
                    f"routing-{router_idx}",
                    service.recip_keys[0].value,
                    PublicKeyType.ED25519_SIG_2018,
                    did_controller,
                    True,
                )
                routing_keys.append(rk)
                svc_endpoints = [service.endpoint]
                break
            router_id = router.inbound_connection_id

        if mediation_records:
            for mediation_record in mediation_records:
                mediator_routing_keys = [
                    PublicKey(
                        did_info.did,
                        f"routing-{idx}",
                        key,
                        PublicKeyType.ED25519_SIG_2018,
                        did_controller,  # TODO: get correct controller did_info
                        True,  # TODO: should this be true?
                    )
                    for idx, key in enumerate(mediation_record.routing_keys)
                ]

                routing_keys = [*routing_keys, *mediator_routing_keys]
                svc_endpoints = [mediation_record.endpoint]

        for (endpoint_index, svc_endpoint) in enumerate(svc_endpoints or []):
            endpoint_ident = "indy" if endpoint_index == 0 else f"indy{endpoint_index}"
            service = Service(
                did_info.did,
                endpoint_ident,
                "IndyAgent",
                [pk],
                routing_keys,
                svc_endpoint,
            )
            did_doc.set(service)

        return did_doc

    async def store_did_document(self, did_doc: DIDDoc):
        """Store a DID document.

        Args:
            did_doc: The `DIDDoc` instance to persist
        """
        assert did_doc.did
        storage: BaseStorage = self._session.inject(BaseStorage)
        try:
            stored_doc, record = await self.fetch_did_document(did_doc.did)
        except StorageNotFoundError:
            record = StorageRecord(
                self.RECORD_TYPE_DID_DOC,
                did_doc.to_json(),
                {"did": did_doc.did},
            )
            await storage.add_record(record)
        else:
            await storage.update_record(record, did_doc.to_json(), {"did": did_doc.did})
        await self.remove_keys_for_did(did_doc.did)
        for key in did_doc.pubkey.values():
            if key.controller == did_doc.did:
                await self.add_key_for_did(did_doc.did, key.value)

    async def add_key_for_did(self, did: str, key: str):
        """Store a verkey for lookup against a DID.

        Args:
            did: The DID to associate with this key
            key: The verkey to be added
        """
        record = StorageRecord(self.RECORD_TYPE_DID_KEY, key, {"did": did, "key": key})
        storage = self._session.inject(BaseStorage)
        await storage.add_record(record)

    async def find_did_for_key(self, key: str) -> str:
        """Find the DID previously associated with a key.

        Args:
            key: The verkey to look up
        """
        storage = self._session.inject(BaseStorage)
        record = await storage.find_record(self.RECORD_TYPE_DID_KEY, {"key": key})
        return record.tags["did"]

    async def remove_keys_for_did(self, did: str):
        """Remove all keys associated with a DID.

        Args:
            did: The DID for which to remove keys
        """
        storage = self._session.inject(BaseStorage)
        await storage.delete_all_records(self.RECORD_TYPE_DID_KEY, {"did": did})

    async def _get_from_ledger(self, public_did: str) -> Tuple[str, Sequence[str]]:
        """Get endpoint and recipient keys for public DID from ledger."""
        ledger = self._session.inject(BaseLedger, required=False)
        if not ledger:
            raise BaseConnectionManagerError(
                f"Cannot resolve DID {public_did} without ledger instance"
            )
        async with ledger:
            endpoint = await ledger.get_endpoint_for_did(public_did)
            recipient_keys = [await ledger.get_key_for_did(public_did)]

        return (endpoint, recipient_keys)

    async def fetch_connection_targets(
        self, connection: ConnRecord
    ) -> Sequence[ConnectionTarget]:
        """Get a list of connection targets from a `ConnRecord`.

        Args:
            connection: The connection record (with associated `DIDDoc`)
                used to generate the connection target
        """

        if not connection.my_did:
            self._logger.debug("No local DID associated with connection")
            return None

        wallet = self._session.inject(BaseWallet)
        my_info = await wallet.get_local_did(connection.my_did)
        results = None

        if (
            ConnRecord.State.get(connection.state)
            in (ConnRecord.State.INVITATION, ConnRecord.State.REQUEST)
            and ConnRecord.Role.get(connection.their_role) is ConnRecord.Role.RESPONDER
        ):
            invitation = await connection.retrieve_invitation(self._session)
            if isinstance(invitation, ConnectionInvitation):  # conn protocol invitation
                if invitation.did:
                    # populate recipient keys, endpoint from ledger
                    (endpoint, recipient_keys) = await self._get_from_ledger(
                        invitation.did
                    )
                    routing_keys = []
                else:
                    endpoint = invitation.endpoint
                    recipient_keys = invitation.recipient_keys
                    routing_keys = invitation.routing_keys
            else:  # out-of-band invitation
                if invitation.service_dids:
                    # populate recipient keys, endpoint from ledger
                    (endpoint, recipient_keys) = await self._get_from_ledger(
                        invitation.service_dids[0]
                    )
                    routing_keys = []
                else:
                    endpoint = invitation.service_blocks[0].service_endpoint
                    recipient_keys = [
                        did_key_to_naked(k)
                        for k in invitation.service_blocks[0].recipient_keys
                    ]
                    routing_keys = [
                        did_key_to_naked(k)
                        for k in invitation.service_blocks[0].routing_keys
                    ]

            results = [
                ConnectionTarget(
                    did=connection.their_did,
                    endpoint=endpoint,
                    label=invitation.label,
                    recipient_keys=recipient_keys,
                    routing_keys=routing_keys,
                    sender_key=my_info.verkey,
                )
            ]
        else:
            if not connection.their_did:
                self._logger.debug("No target DID associated with connection")
                return None

            did_doc, _ = await self.fetch_did_document(connection.their_did)
            results = self.diddoc_connection_targets(
                did_doc, my_info.verkey, connection.their_label
            )

        return results

    def diddoc_connection_targets(
        self, doc: DIDDoc, sender_verkey: str, their_label: str = None
    ) -> Sequence[ConnectionTarget]:
        """Get a list of connection targets from a DID Document.

        Args:
            doc: The DID Document to create the target from
            sender_verkey: The verkey we are using
            their_label: The connection label they are using
        """

        if not doc:
            raise BaseConnectionManagerError("No DIDDoc provided for connection target")
        if not doc.did:
            raise BaseConnectionManagerError("DIDDoc has no DID")
        if not doc.service:
            raise BaseConnectionManagerError("No services defined by DIDDoc")

        targets = []
        for service in doc.service.values():
            if service.recip_keys:
                targets.append(
                    ConnectionTarget(
                        did=doc.did,
                        endpoint=service.endpoint,
                        label=their_label,
                        recipient_keys=[
                            key.value for key in (service.recip_keys or ())
                        ],
                        routing_keys=[
                            key.value for key in (service.routing_keys or ())
                        ],
                        sender_key=sender_verkey,
                    )
                )
        return targets

    async def fetch_did_document(self, did: str) -> Tuple[DIDDoc, StorageRecord]:
        """Retrieve a DID Document for a given DID.

        Args:
            did: The DID to search for
        """
        storage = self._session.inject(BaseStorage)
        record = await storage.find_record(self.RECORD_TYPE_DID_DOC, {"did": did})
        return DIDDoc.from_json(record.value), record



"""Credential issue message."""

from typing import Sequence

from marshmallow import EXCLUDE, fields

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from .....messaging.decorators.attach_decorator import (
    AttachDecorator,
    AttachDecoratorSchema,
)
from .....messaging.valid import UUIDFour

from ..message_types import CRED_20_ISSUE, PROTOCOL_PACKAGE

from .cred_format import V20CredFormat, V20CredFormatSchema

HANDLER_CLASS = f"{PROTOCOL_PACKAGE}.handlers.cred_issue_handler.V20CredIssueHandler"


class V20CredIssue(AgentMessage):
    """Credential issue message."""

    class Meta:
        """V20CredIssue metadata."""

        handler_class = HANDLER_CLASS
        schema_class = "V20CredIssueSchema"
        message_type = CRED_20_ISSUE

    def __init__(
        self,
        _id: str = None,
        *,
        replacement_id: str = None,
        comment: str = None,
        formats: Sequence[V20CredFormat] = None,
        credentials_attach: Sequence[AttachDecorator] = None,
        **kwargs,
    ):
        """
        Initialize credential issue object.

        Args:
            comment: optional comment
            credentials_attach: credentials attachments

        """
        super().__init__(_id=_id, **kwargs)
        self.replacement_id = replacement_id
        self.comment = comment
        self.formats = list(formats) if formats else []
        self.credentials_attach = list(credentials_attach) if credentials_attach else []

    def cred(self, fmt: V20CredFormat.Format = None) -> dict:
        """
        Return attached credential.

        Args:
            fmt: format of attachment in list to decode and return

        """
        return (fmt or V20CredFormat.Format.INDY).get_attachment_data(
            self.formats,
            self.credentials_attach,
        )


class V20CredIssueSchema(AgentMessageSchema):
    """Credential issue schema."""

    class Meta:
        """Credential issue schema metadata."""

        model_class = V20CredIssue
        unknown = EXCLUDE

    replacement_id = fields.Str(
        description="Issuer-unique identifier to coordinate credential replacement",
        required=False,
        allow_none=False,
        example=UUIDFour.EXAMPLE,
    )
    comment = fields.Str(
        description="Human-readable comment", required=False, allow_none=True
    )
    formats = fields.Nested(
        V20CredFormatSchema,
        many=True,
        required=True,
        description="Acceptable credential formats",
    )
    credentials_attach = fields.Nested(
        AttachDecoratorSchema, many=True, required=True, data_key="credentials~attach"
    )



import asyncio

from asynctest import mock as async_mock, TestCase as AsyncTestCase

from .. import basic as test_module
from ..basic import BasicMessageQueue


async def collect(queue, count=1):
    found = []
    async for result in queue:
        found.append(result)
        if len(found) >= count:
            queue.stop()
    return found


class TestBasicQueue(AsyncTestCase):
    async def test_enqueue_dequeue(self):
        queue = BasicMessageQueue()

        with self.assertRaises(asyncio.TimeoutError):
            await queue.dequeue(timeout=0)

        test_value = "test value"
        await queue.enqueue(test_value)
        assert await queue.dequeue(timeout=0) == test_value
        with self.assertRaises(asyncio.TimeoutError):
            await queue.dequeue(timeout=0)

        queue.task_done()
        await queue.join()

    async def test_dequeue_x(self):
        queue = BasicMessageQueue()
        test_value = "test value"
        await queue.enqueue(test_value)

        with async_mock.patch.object(
            test_module.asyncio, "get_event_loop", async_mock.MagicMock()
        ) as mock_get_event_loop, async_mock.patch.object(
            test_module.asyncio, "wait", async_mock.CoroutineMock()
        ) as mock_wait:
            mock_wait.return_value = (
                async_mock.MagicMock(),
                [
                    async_mock.MagicMock(
                        done=async_mock.MagicMock(), cancel=async_mock.MagicMock()
                    )
                ],
            )
            mock_get_event_loop.return_value = async_mock.MagicMock(
                create_task=async_mock.MagicMock(
                    side_effect=[
                        async_mock.MagicMock(),  # stopped
                        async_mock.MagicMock(  # dequeued
                            done=async_mock.MagicMock(return_value=True),
                            exception=async_mock.MagicMock(return_value=KeyError()),
                        ),
                    ]
                )
            )
            with self.assertRaises(KeyError):
                await queue.dequeue(timeout=0)

    async def test_dequeue_none(self):
        queue = BasicMessageQueue()
        test_value = "test value"
        await queue.enqueue(test_value)

        with async_mock.patch.object(
            test_module.asyncio, "get_event_loop", async_mock.MagicMock()
        ) as mock_get_event_loop, async_mock.patch.object(
            test_module.asyncio, "wait", async_mock.CoroutineMock()
        ) as mock_wait:
            mock_wait.return_value = (
                async_mock.MagicMock(),
                [
                    async_mock.MagicMock(
                        done=async_mock.MagicMock(), cancel=async_mock.MagicMock()
                    )
                ],
            )
            mock_get_event_loop.return_value = async_mock.MagicMock(
                create_task=async_mock.MagicMock(
                    side_effect=[
                        async_mock.MagicMock(  # stopped
                            done=async_mock.MagicMock(return_value=True)
                        ),
                        async_mock.MagicMock(  # dequeued
                            done=async_mock.MagicMock(return_value=False)
                        ),
                    ]
                )
            )
            assert await queue.dequeue(timeout=0) is None

    async def test_async_iter(self):
        queue = BasicMessageQueue()

        results = asyncio.wait_for(collect(queue), timeout=1.0)
        test_value = "test value"
        await queue.enqueue(test_value)
        found = await results
        assert found == [test_value]

    async def test_stopped(self):
        queue = BasicMessageQueue()
        queue.stop()

        with self.assertRaises(asyncio.CancelledError):
            await queue.dequeue(timeout=0)

        test_value = "test value"
        with self.assertRaises(asyncio.CancelledError):
            await queue.enqueue(test_value)
        results = asyncio.wait_for(collect(queue), timeout=1.0)
        assert await results == []
        with self.assertRaises(asyncio.CancelledError):
            await queue.dequeue(timeout=0)

        queue.reset()
        results = asyncio.wait_for(collect(queue), timeout=1.0)
        await queue.enqueue(test_value)
        found = await results
        assert found == [test_value]



from unittest import mock, TestCase

from .....didcomm_prefix import DIDCommPrefix

from ...message_types import PROTOCOL_PACKAGE, ROUTE_QUERY_REQUEST
from ...models.paginate import Paginate, PaginateSchema

from ..route_query_request import RouteQueryRequest


class TestRouteQueryRequest(TestCase):
    test_limit = 100
    test_offset = 10
    test_verkey = "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"
    test_filter = {"recipient_key": ["3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"]}

    def setUp(self):
        self.paginate = Paginate(limit=self.test_limit, offset=self.test_offset)
        self.message = RouteQueryRequest(
            filter=self.test_filter, paginate=self.paginate
        )

    def test_init(self):
        assert self.message.filter == self.test_filter
        assert self.message.paginate.limit == self.test_limit
        assert self.message.paginate.offset == self.test_offset

    def test_type(self):
        assert self.message._type == DIDCommPrefix.qualify_current(ROUTE_QUERY_REQUEST)

    @mock.patch(
        f"{PROTOCOL_PACKAGE}.messages.route_query_request.RouteQueryRequestSchema.load"
    )
    def test_deserialize(self, message_schema_load):
        obj = {"obj": "obj"}

        message = RouteQueryRequest.deserialize(obj)
        message_schema_load.assert_called_once_with(obj)

        assert message is message_schema_load.return_value

    @mock.patch(
        f"{PROTOCOL_PACKAGE}.messages.route_query_request.RouteQueryRequestSchema.dump"
    )
    def test_serialize(self, message_schema_dump):
        message_dict = self.message.serialize()
        message_schema_dump.assert_called_once_with(self.message)

        assert message_dict is message_schema_dump.return_value


class TestRouteQueryRequestSchema(TestCase):
    def test_make_model(self):
        message = RouteQueryRequest(filter={}, paginate=Paginate())
        data = message.serialize()
        model_instance = RouteQueryRequest.deserialize(data)
        assert isinstance(model_instance, RouteQueryRequest)



"""Handler for keylist-update messages."""

from .....messaging.base_handler import (
    BaseHandler,
    BaseResponder,
    HandlerException,
    RequestContext,
)
from .....storage.error import StorageNotFoundError
from ....problem_report.v1_0.message import ProblemReport
from ..manager import MediationManager, MediationNotGrantedError
from ..messages.keylist_update import KeylistUpdate
from ..models.mediation_record import MediationRecord


class KeylistUpdateHandler(BaseHandler):
    """Handler for keylist-update messages."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Handle keylist-update messages."""
        self._logger.debug(
            "%s called with context %s", self.__class__.__name__, context
        )
        assert isinstance(context.message, KeylistUpdate)

        if not context.connection_ready:
            raise HandlerException("Cannot update routes: no active connection")

        mgr = MediationManager(context.profile)
        try:
            async with context.session() as session:
                record = await MediationRecord.retrieve_by_connection_id(
                    session, context.connection_record.connection_id
                )
            response = await mgr.update_keylist(record, updates=context.message.updates)
            await responder.send_reply(response)
        except (StorageNotFoundError, MediationNotGrantedError):
            await responder.send_reply(
                ProblemReport(
                    explain_ltxt="Mediation has not been granted for this connection."
                )
            )



"""
A message decorator for trace events.

A trace decorator identifies a responsibility on the processor
to record information on message processing events.
"""

from typing import Sequence

from marshmallow import EXCLUDE, fields

from ..models.base import BaseModel, BaseModelSchema
from ..valid import UUIDFour


TRACE_MESSAGE_TARGET = "message"
TRACE_LOG_TARGET = "log"


class TraceReport(BaseModel):
    """Class representing a Trace Report."""

    class Meta:
        """TraceReport metadata."""

        schema_class = "TraceReport"

    def __init__(
        self,
        *,
        msg_id: str = None,
        thread_id: str = None,
        traced_type: str = None,
        timestamp: str = None,
        str_time: str = None,
        handler: str = None,
        ellapsed_milli: int = None,
        outcome: str = None,
    ):
        """
        Initialize a TraceReport instance.

        Args:
            msg_id: ...
            thread_id: ...
            traced_type: ...
            timestamp: ...
            str_time: ...
            handler: ...
            ellapsed_milli: ...
            outcome: ...
        """
        super().__init__()
        self._msg_id = msg_id
        self._thread_id = thread_id
        self._traced_type = traced_type
        self._timestamp = timestamp
        self._str_time = str_time
        self._handler = handler
        self._ellapsed_milli = ellapsed_milli
        self._outcome = outcome

    @property
    def msg_id(self):
        """
        Accessor for msg_id.

        Returns:
            The msg_id

        """
        return self._msg_id

    @property
    def thread_id(self):
        """
        Accessor for thread_id.

        Returns:
            The thread_id

        """
        return self._thread_id

    @property
    def traced_type(self):
        """
        Accessor for traced_type.

        Returns:
            The sender traced_type

        """
        return self._traced_type

    @property
    def timestamp(self):
        """
        Accessor for timestamp.

        Returns:
            The sender timestamp

        """
        return self._timestamp

    @property
    def str_time(self):
        """
        Accessor for str_time.

        Returns:
            Formatted representation of the sender timestamp

        """
        return self._str_time

    @property
    def handler(self):
        """
        Accessor for handler.

        Returns:
            The sender handler

        """
        return self._handler

    @property
    def ellapsed_milli(self):
        """
        Accessor for ellapsed_milli.

        Returns:
            The sender ellapsed_milli

        """
        return self._ellapsed_milli

    @property
    def outcome(self):
        """
        Accessor for outcome.

        Returns:
            The sender outcome

        """
        return self._outcome


class TraceDecorator(BaseModel):
    """Class representing trace decorator."""

    class Meta:
        """TraceDecorator metadata."""

        schema_class = "TraceDecoratorSchema"

    def __init__(
        self,
        *,
        target: str = None,
        full_thread: bool = True,
        trace_reports: Sequence = None,
    ):
        """
        Initialize a TraceDecorator instance.

        Args:
            target: The "target" can refer to a url (as above) or the term
                    "message", which is a request to append trace information
                    to the message itself.
            full_thread: An optional flag to indicate tracing should be included
                    on all subsequent messages in the thread (on by default).
            trace_reports: Trace reports contain information about a message
                    processing at a specific point in time, along with a timestamp.
                    Trace reports can be used to identify steps in the processing
                    of a message or thread, and support troubleshooting and
                    performance issues.
        """
        super(TraceDecorator, self).__init__()
        self._target = target
        self._full_thread = full_thread
        self._trace_reports = trace_reports and list(trace_reports) or None

    @property
    def target(self):
        """
        Accessor for trace target.

        Returns:
            The target for tracing messages

        """
        return self._target

    @property
    def full_thread(self):
        """
        Accessor for full_thread flag.

        Returns:
            The full_thread flag

        """
        return self._full_thread

    @property
    def trace_reports(self):
        """
        Set of trace reports for this message.

        Returns:
            The trace reports that have been logged on this message/thread
            so far.  (Only for target="message".)

        """
        if not self._trace_reports:
            return []
        return self._trace_reports

    def append_trace_report(self, trace_report: TraceReport):
        """Append a trace report to this decorator."""
        if not self._trace_reports:
            self._trace_reports = []
        self._trace_reports.append(trace_report)


class TraceReportSchema(BaseModelSchema):
    """Trace report schema."""

    class Meta:
        """TraceReportSchema metadata."""

        model_class = TraceReport
        unknown = EXCLUDE

    msg_id = fields.Str(
        required=True,
        allow_none=False,
        description="Message Id",
        example=UUIDFour.EXAMPLE,  # typically a UUID4 but not necessarily
    )
    thread_id = fields.Str(
        required=True,
        allow_none=False,
        description="Message Id",
        example=UUIDFour.EXAMPLE,  # typically a UUID4 but not necessarily
    )
    traced_type = fields.Str(
        required=False,
        allow_none=True,
        description="Type of traced message",
        example="TODO",
    )
    timestamp = fields.Str(
        required=True,
        allow_none=False,
        description="Timestamp of traced event",
        example="123456789.123456",
    )
    str_time = fields.Str(
        required=True,
        allow_none=False,
        description="Formatted timestamp of traced event",
        example="2018-03-27 18:23:45.123Z",
    )
    handler = fields.Str(
        required=False,
        allow_none=True,
        description="Description of the message handler",
        example="TODO",
    )
    ellapsed_milli = fields.Int(
        required=False,
        allow_none=True,
        description="Elapsed milliseconds processing time",
        example=27,
        strict=True,
    )
    outcome = fields.Str(
        required=False,
        allow_none=True,
        description="Outcome description",
        example="TODO",
    )


class TraceDecoratorSchema(BaseModelSchema):
    """Trace decorator schema used in serialization/deserialization."""

    class Meta:
        """TraceDecoratorSchema metadata."""

        model_class = TraceDecorator
        unknown = EXCLUDE

    target = fields.Str(
        required=True,
        allow_none=False,
        description="Trace report target",
        example="'http://example.com/tracer', or 'message'",
    )
    full_thread = fields.Boolean(
        required=False,
        allow_none=True,
        description="Parent thread identifier",
        example="True",
    )
    trace_reports = fields.List(
        fields.Nested(TraceReportSchema),
        required=False,
        allow_none=True,
        description=("The set of reports collected so far for this message or thread"),
    )



"""Response to keylist-update used to notify mediation client of applied updates."""

from typing import Sequence

from marshmallow import fields

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from ..message_types import KEYLIST_UPDATE_RESPONSE, PROTOCOL_PACKAGE

from .inner.keylist_updated import KeylistUpdated, KeylistUpdatedSchema

HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers"
    ".keylist_update_response_handler.KeylistUpdateResponseHandler"
)


class KeylistUpdateResponse(AgentMessage):
    """Class representing a keylist update result message."""

    class Meta:
        """Metadata for a keylist update result."""

        handler_class = HANDLER_CLASS
        message_type = KEYLIST_UPDATE_RESPONSE
        schema_class = "KeylistUpdateResponseSchema"

    def __init__(
        self,
        *,
        updated: Sequence[KeylistUpdated] = None,
        **kwargs,
    ):
        """
        Initialize keylist update object.

        Args:
            updates: Update rules for keylist update request
        """
        super(KeylistUpdateResponse, self).__init__(**kwargs)
        self.updated = list(updated) if updated else []


class KeylistUpdateResponseSchema(AgentMessageSchema):
    """Keylist update result schema class."""

    class Meta:
        """Keylist update result schema metadata."""

        model_class = KeylistUpdateResponse

    updated = fields.List(
        fields.Nested(KeylistUpdatedSchema()),
        description="List of update results for each update",
    )



"""Handle registration and publication of supported protocols."""

import logging

from typing import Mapping, Sequence

from ..config.injection_context import InjectionContext
from ..utils.classloader import ClassLoader

from .error import ProtocolMinorVersionNotSupported

LOGGER = logging.getLogger(__name__)


class ProtocolRegistry:
    """Protocol registry for indexing message families."""

    def __init__(self):
        """Initialize a `ProtocolRegistry` instance."""
        self._controllers = {}
        self._typemap = {}
        self._versionmap = {}

    @property
    def protocols(self) -> Sequence[str]:
        """Accessor for a list of all message protocols."""
        prots = set()
        for message_type in self._typemap.keys():
            pos = message_type.rfind("/")
            if pos > 0:
                family = message_type[:pos]
                prots.add(family)
        return prots

    @property
    def message_types(self) -> Sequence[str]:
        """Accessor for a list of all message types."""
        return tuple(self._typemap.keys())

    @property
    def controllers(self) -> Mapping[str, str]:
        """Accessor for a list of all protocol controller functions."""
        return self._controllers.copy()

    def protocols_matching_query(self, query: str) -> Sequence[str]:
        """Return a list of message protocols matching a query string."""
        all_types = self.protocols
        result = None

        if query == "*" or query is None:
            result = all_types
        elif query:
            if query.endswith("*"):
                match = query[:-1]
                result = tuple(k for k in all_types if k.startswith(match))
            elif query in all_types:
                result = (query,)
        return result or ()

    def parse_type_string(self, message_type):
        """Parse message type string and return dict with info."""
        tokens = message_type.split("/")
        protocol_name = tokens[-3]
        version_string = tokens[-2]
        message_name = tokens[-1]

        version_string_tokens = version_string.split(".")
        assert len(version_string_tokens) == 2

        return {
            "protocol_name": protocol_name,
            "message_name": message_name,
            "major_version": int(version_string_tokens[0]),
            "minor_version": int(version_string_tokens[1]),
        }

    def register_message_types(self, *typesets, version_definition=None):
        """
        Add new supported message types.

        Args:
            typesets: Mappings of message types to register
            version_definition: Optional version definition dict

        """

        # Maintain support for versionless protocol modules
        for typeset in typesets:
            self._typemap.update(typeset)

        # Track versioned modules for version routing
        if version_definition:
            for typeset in typesets:
                for message_type_string, module_path in typeset.items():
                    parsed_type_string = self.parse_type_string(message_type_string)

                    if version_definition["major_version"] not in self._versionmap:
                        self._versionmap[version_definition["major_version"]] = []

                    self._versionmap[version_definition["major_version"]].append(
                        {
                            "parsed_type_string": parsed_type_string,
                            "version_definition": version_definition,
                            "message_module": module_path,
                        }
                    )

    def register_controllers(self, *controller_sets, version_definition=None):
        """
        Add new controllers.

        Args:
            controller_sets: Mappings of message families to coroutines

        """
        for controlset in controller_sets:
            self._controllers.update(controlset)

    def resolve_message_class(self, message_type: str) -> type:
        """
        Resolve a message_type to a message class.

        Given a message type identifier, this method
        returns the corresponding registered message class.

        Args:
            message_type: Message type to resolve

        Returns:
            The resolved message class

        """

        # Try and retrieve from direct mapping
        msg_cls = self._typemap.get(message_type)
        if isinstance(msg_cls, str):
            return ClassLoader.load_class(msg_cls)

        # Support registered modules (not path as string)
        elif msg_cls:
            return msg_cls

        # Try and route via min/maj version matching
        if not msg_cls:
            parsed_type_string = self.parse_type_string(message_type)
            major_version = parsed_type_string["major_version"]

            version_supported_protos = self._versionmap.get(major_version)
            if not version_supported_protos:
                return None

            for proto in version_supported_protos:
                if (
                    proto["parsed_type_string"]["protocol_name"]
                    == parsed_type_string["protocol_name"]
                    and proto["parsed_type_string"]["message_name"]
                    == parsed_type_string["message_name"]
                ):
                    if (
                        parsed_type_string["minor_version"]
                        < proto["version_definition"]["minimum_minor_version"]
                    ):
                        raise ProtocolMinorVersionNotSupported(
                            "Minimum supported minor version is "
                            + f"{proto['version_definition']['minimum_minor_version']}."
                            + f" Received {parsed_type_string['minor_version']}."
                        )

                    if isinstance(proto["message_module"], str):
                        return ClassLoader.load_class(msg_cls)
                    elif proto["message_module"]:
                        return proto["message_module"]

        return None

    async def prepare_disclosed(
        self, context: InjectionContext, protocols: Sequence[str]
    ):
        """Call controllers and return publicly supported message families and roles."""
        published = []
        for protocol in protocols:
            result = {"pid": protocol}
            if protocol in self._controllers:
                ctl_cls = self._controllers[protocol]
                if isinstance(ctl_cls, str):
                    ctl_cls = ClassLoader.load_class(ctl_cls)
                ctl_instance = ctl_cls(protocol)
                if hasattr(ctl_instance, "check_access"):
                    allowed = await ctl_instance.check_access(context)
                    if not allowed:
                        # remove from published
                        continue
                if hasattr(ctl_instance, "determine_roles"):
                    roles = await ctl_instance.determine_roles(context)
                    if roles:
                        result["roles"] = list(roles)
            published.append(result)
        return published

    def __repr__(self) -> str:
        """Return a string representation for this class."""
        return "<{}>".format(self.__class__.__name__)



"""Classes to manage credentials."""

import asyncio
import json
import logging

from typing import Mapping, Tuple

from ....cache.base import BaseCache
from ....core.error import BaseError
from ....core.profile import Profile
from ....indy.holder import IndyHolder, IndyHolderError
from ....indy.issuer import IndyIssuer, IndyIssuerRevocationRegistryFullError
from ....ledger.base import BaseLedger
from ....messaging.credential_definitions.util import (
    CRED_DEF_TAGS,
    CRED_DEF_SENT_RECORD_TYPE,
)
from ....revocation.indy import IndyRevocation
from ....revocation.models.revocation_registry import RevocationRegistry
from ....revocation.models.issuer_rev_reg_record import IssuerRevRegRecord
from ....storage.base import BaseStorage
from ....storage.error import StorageNotFoundError

from .messages.credential_ack import CredentialAck
from .messages.credential_issue import CredentialIssue
from .messages.credential_offer import CredentialOffer
from .messages.credential_proposal import CredentialProposal
from .messages.credential_request import CredentialRequest
from .messages.inner.credential_preview import CredentialPreview
from .models.credential_exchange import V10CredentialExchange

LOGGER = logging.getLogger(__name__)


class CredentialManagerError(BaseError):
    """Credential error."""


class CredentialManager:
    """Class for managing credentials."""

    def __init__(self, profile: Profile):
        """
        Initialize a CredentialManager.

        Args:
            profile: The profile instance for this credential manager
        """
        self._profile = profile

    @property
    def profile(self) -> Profile:
        """
        Accessor for the current profile instance.

        Returns:
            The profile instance for this credential manager

        """
        return self._profile

    async def _match_sent_cred_def_id(self, tag_query: Mapping[str, str]) -> str:
        """Return most recent matching id of cred def that agent sent to ledger."""

        async with self._profile.session() as session:
            storage = session.inject(BaseStorage)
            found = await storage.find_all_records(
                type_filter=CRED_DEF_SENT_RECORD_TYPE, tag_query=tag_query
            )
        if not found:
            raise CredentialManagerError(
                f"Issuer has no operable cred def for proposal spec {tag_query}"
            )
        return max(found, key=lambda r: int(r.tags["epoch"])).tags["cred_def_id"]

    async def prepare_send(
        self,
        connection_id: str,
        credential_proposal: CredentialProposal,
        auto_remove: bool = None,
    ) -> Tuple[V10CredentialExchange, CredentialOffer]:
        """
        Set up a new credential exchange for an automated send.

        Args:
            connection_id: Connection to create offer for
            credential_proposal: The credential proposal with preview
            auto_remove: Flag to automatically remove the record on completion

        Returns:
            A tuple of the new credential exchange record and credential offer message

        """
        if auto_remove is None:
            auto_remove = not self._profile.settings.get("preserve_exchange_records")
        credential_exchange = V10CredentialExchange(
            connection_id=connection_id,
            initiator=V10CredentialExchange.INITIATOR_SELF,
            role=V10CredentialExchange.ROLE_ISSUER,
            credential_proposal_dict=credential_proposal.serialize(),
            auto_issue=True,
            auto_remove=auto_remove,
            trace=(credential_proposal._trace is not None),
        )
        (credential_exchange, credential_offer) = await self.create_offer(
            cred_ex_record=credential_exchange,
            comment="create automated credential exchange",
        )
        return (credential_exchange, credential_offer)

    async def create_proposal(
        self,
        connection_id: str,
        *,
        auto_offer: bool = None,
        auto_remove: bool = None,
        comment: str = None,
        credential_preview: CredentialPreview = None,
        schema_id: str = None,
        schema_issuer_did: str = None,
        schema_name: str = None,
        schema_version: str = None,
        cred_def_id: str = None,
        issuer_did: str = None,
        trace: bool = False,
    ) -> V10CredentialExchange:
        """
        Create a credential proposal.

        Args:
            connection_id: Connection to create proposal for
            auto_offer: Should this proposal request automatically be handled to
                offer a credential
            auto_remove: Should the record be automatically removed on completion
            comment: Optional human-readable comment to include in proposal
            credential_preview: The credential preview to use to create
                the credential proposal
            schema_id: Schema id for credential proposal
            schema_issuer_did: Schema issuer DID for credential proposal
            schema_name: Schema name for credential proposal
            schema_version: Schema version for credential proposal
            cred_def_id: Credential definition id for credential proposal
            issuer_did: Issuer DID for credential proposal

        Returns:
            Resulting credential exchange record including credential proposal

        """
        credential_proposal_message = CredentialProposal(
            comment=comment,
            credential_proposal=credential_preview,
            schema_id=schema_id,
            schema_issuer_did=schema_issuer_did,
            schema_name=schema_name,
            schema_version=schema_version,
            cred_def_id=cred_def_id,
            issuer_did=issuer_did,
        )
        credential_proposal_message.assign_trace_decorator(
            self._profile.settings, trace
        )

        if auto_remove is None:
            auto_remove = not self._profile.settings.get("preserve_exchange_records")
        cred_ex_record = V10CredentialExchange(
            connection_id=connection_id,
            thread_id=credential_proposal_message._thread_id,
            initiator=V10CredentialExchange.INITIATOR_SELF,
            role=V10CredentialExchange.ROLE_HOLDER,
            state=V10CredentialExchange.STATE_PROPOSAL_SENT,
            credential_proposal_dict=credential_proposal_message.serialize(),
            auto_offer=auto_offer,
            auto_remove=auto_remove,
            trace=trace,
        )
        async with self._profile.session() as session:
            await cred_ex_record.save(session, reason="create credential proposal")
        return cred_ex_record

    async def receive_proposal(
        self, message: CredentialProposal, connection_id: str
    ) -> V10CredentialExchange:
        """
        Receive a credential proposal.

        Returns:
            The resulting credential exchange record, created

        """
        # at this point, cred def and schema still open to potential negotiation
        cred_ex_record = V10CredentialExchange(
            connection_id=connection_id,
            thread_id=message._thread_id,
            initiator=V10CredentialExchange.INITIATOR_EXTERNAL,
            role=V10CredentialExchange.ROLE_ISSUER,
            state=V10CredentialExchange.STATE_PROPOSAL_RECEIVED,
            credential_proposal_dict=message.serialize(),
            auto_offer=self._profile.settings.get(
                "debug.auto_respond_credential_proposal"
            ),
            auto_issue=self._profile.settings.get(
                "debug.auto_respond_credential_request"
            ),
            auto_remove=not self._profile.settings.get("preserve_exchange_records"),
            trace=(message._trace is not None),
        )
        async with self._profile.session() as session:
            await cred_ex_record.save(session, reason="receive credential proposal")

        return cred_ex_record

    async def create_offer(
        self, cred_ex_record: V10CredentialExchange, comment: str = None
    ) -> Tuple[V10CredentialExchange, CredentialOffer]:
        """
        Create a credential offer, update credential exchange record.

        Args:
            cred_ex_record: Credential exchange to create offer for
            comment: optional human-readable comment to set in offer message

        Returns:
            A tuple (credential exchange record, credential offer message)

        """

        async def _create(cred_def_id):
            issuer = self._profile.inject(IndyIssuer)
            offer_json = await issuer.create_credential_offer(cred_def_id)
            return json.loads(offer_json)

        credential_proposal_message = CredentialProposal.deserialize(
            cred_ex_record.credential_proposal_dict
        )
        credential_proposal_message.assign_trace_decorator(
            self._profile.settings, cred_ex_record.trace
        )
        cred_def_id = await self._match_sent_cred_def_id(
            {
                t: getattr(credential_proposal_message, t)
                for t in CRED_DEF_TAGS
                if getattr(credential_proposal_message, t)
            }
        )
        cred_preview = credential_proposal_message.credential_proposal

        # vet attributes
        ledger = self._profile.inject(BaseLedger)
        async with ledger:
            schema_id = await ledger.credential_definition_id2schema_id(cred_def_id)
            schema = await ledger.get_schema(schema_id)
        schema_attrs = {attr for attr in schema["attrNames"]}
        preview_attrs = {attr for attr in cred_preview.attr_dict()}
        if preview_attrs != schema_attrs:
            raise CredentialManagerError(
                f"Preview attributes {preview_attrs} "
                f"mismatch corresponding schema attributes {schema_attrs}"
            )

        credential_offer = None
        cache_key = f"credential_offer::{cred_def_id}"
        cache = self._profile.inject(BaseCache, required=False)
        if cache:
            async with cache.acquire(cache_key) as entry:
                if entry.result:
                    credential_offer = entry.result
                else:
                    credential_offer = await _create(cred_def_id)
                    await entry.set_result(credential_offer, 3600)
        if not credential_offer:
            credential_offer = await _create(cred_def_id)

        credential_offer_message = CredentialOffer(
            comment=comment,
            credential_preview=cred_preview,
            offers_attach=[CredentialOffer.wrap_indy_offer(credential_offer)],
        )

        credential_offer_message._thread = {"thid": cred_ex_record.thread_id}
        credential_offer_message.assign_trace_decorator(
            self._profile.settings, cred_ex_record.trace
        )

        cred_ex_record.thread_id = credential_offer_message._thread_id
        cred_ex_record.schema_id = credential_offer["schema_id"]
        cred_ex_record.credential_definition_id = credential_offer["cred_def_id"]
        cred_ex_record.state = V10CredentialExchange.STATE_OFFER_SENT
        cred_ex_record.credential_offer = credential_offer

        cred_ex_record.credential_offer_dict = credential_offer_message.serialize()

        async with self._profile.session() as session:
            await cred_ex_record.save(session, reason="create credential offer")

        return (cred_ex_record, credential_offer_message)

    async def receive_offer(
        self, message: CredentialOffer, connection_id: str
    ) -> V10CredentialExchange:
        """
        Receive a credential offer.

        Returns:
            The credential exchange record, updated

        """
        credential_preview = message.credential_preview
        indy_offer = message.indy_offer(0)
        schema_id = indy_offer["schema_id"]
        cred_def_id = indy_offer["cred_def_id"]

        credential_proposal_dict = CredentialProposal(
            comment=message.comment,
            credential_proposal=credential_preview,
            schema_id=schema_id,
            cred_def_id=cred_def_id,
        ).serialize()

        async with self._profile.session() as session:
            # Get credential exchange record (holder sent proposal first)
            # or create it (issuer sent offer first)
            try:
                cred_ex_record = await (
                    V10CredentialExchange.retrieve_by_connection_and_thread(
                        session, connection_id, message._thread_id
                    )
                )
                cred_ex_record.credential_proposal_dict = credential_proposal_dict
            except StorageNotFoundError:  # issuer sent this offer free of any proposal
                cred_ex_record = V10CredentialExchange(
                    connection_id=connection_id,
                    thread_id=message._thread_id,
                    initiator=V10CredentialExchange.INITIATOR_EXTERNAL,
                    role=V10CredentialExchange.ROLE_HOLDER,
                    credential_proposal_dict=credential_proposal_dict,
                    auto_remove=not self._profile.settings.get(
                        "preserve_exchange_records"
                    ),
                    trace=(message._trace is not None),
                )

            cred_ex_record.credential_offer = indy_offer
            cred_ex_record.state = V10CredentialExchange.STATE_OFFER_RECEIVED
            cred_ex_record.schema_id = schema_id
            cred_ex_record.credential_definition_id = cred_def_id

            await cred_ex_record.save(session, reason="receive credential offer")

        return cred_ex_record

    async def create_request(
        self, cred_ex_record: V10CredentialExchange, holder_did: str
    ) -> Tuple[V10CredentialExchange, CredentialRequest]:
        """
        Create a credential request.

        Args:
            cred_ex_record: Credential exchange record
                for which to create request
            holder_did: holder DID

        Returns:
            A tuple (credential exchange record, credential request message)

        """
        if cred_ex_record.state != V10CredentialExchange.STATE_OFFER_RECEIVED:
            raise CredentialManagerError(
                f"Credential exchange {cred_ex_record.credential_exchange_id} "
                f"in {cred_ex_record.state} state "
                f"(must be {V10CredentialExchange.STATE_OFFER_RECEIVED})"
            )

        credential_definition_id = cred_ex_record.credential_definition_id
        credential_offer = cred_ex_record.credential_offer

        async def _create():
            ledger = self._profile.inject(BaseLedger)
            async with ledger:
                credential_definition = await ledger.get_credential_definition(
                    credential_definition_id
                )

            holder = self._profile.inject(IndyHolder)
            request_json, metadata_json = await holder.create_credential_request(
                credential_offer, credential_definition, holder_did
            )
            return {
                "request": json.loads(request_json),
                "metadata": json.loads(metadata_json),
            }

        if cred_ex_record.credential_request:
            LOGGER.warning(
                "create_request called multiple times for v1.0 credential exchange: %s",
                cred_ex_record.credential_exchange_id,
            )
        else:
            if "nonce" not in credential_offer:
                raise CredentialManagerError("Missing nonce in credential offer")
            nonce = credential_offer["nonce"]
            cache_key = (
                f"credential_request::{credential_definition_id}::{holder_did}::{nonce}"
            )
            cred_req_result = None
            cache = self._profile.inject(BaseCache, required=False)
            if cache:
                async with cache.acquire(cache_key) as entry:
                    if entry.result:
                        cred_req_result = entry.result
                    else:
                        cred_req_result = await _create()
                        await entry.set_result(cred_req_result, 3600)
            if not cred_req_result:
                cred_req_result = await _create()

            (
                cred_ex_record.credential_request,
                cred_ex_record.credential_request_metadata,
            ) = (cred_req_result["request"], cred_req_result["metadata"])

        credential_request_message = CredentialRequest(
            requests_attach=[
                CredentialRequest.wrap_indy_cred_req(cred_ex_record.credential_request)
            ]
        )
        credential_request_message._thread = {"thid": cred_ex_record.thread_id}
        credential_request_message.assign_trace_decorator(
            self._profile.settings, cred_ex_record.trace
        )

        cred_ex_record.state = V10CredentialExchange.STATE_REQUEST_SENT
        async with self._profile.session() as session:
            await cred_ex_record.save(session, reason="create credential request")

        return (cred_ex_record, credential_request_message)

    async def receive_request(self, message: CredentialRequest, connection_id: str):
        """
        Receive a credential request.

        Args:
            credential_request_message: Credential request to receive

        Returns:
            credential exchange record, retrieved and updated

        """
        assert len(message.requests_attach or []) == 1
        credential_request = message.indy_cred_req(0)

        async with self._profile.session() as session:
            cred_ex_record = await (
                V10CredentialExchange.retrieve_by_connection_and_thread(
                    session, connection_id, message._thread_id
                )
            )
            cred_ex_record.credential_request = credential_request
            cred_ex_record.state = V10CredentialExchange.STATE_REQUEST_RECEIVED
            await cred_ex_record.save(session, reason="receive credential request")

        return cred_ex_record

    async def issue_credential(
        self,
        cred_ex_record: V10CredentialExchange,
        *,
        comment: str = None,
        retries: int = 5,
    ) -> Tuple[V10CredentialExchange, CredentialIssue]:
        """
        Issue a credential.

        Args:
            cred_ex_record: The credential exchange record
                for which to issue a credential
            comment: optional human-readable comment pertaining to credential issue

        Returns:
            Tuple: (Updated credential exchange record, credential message)

        """

        if cred_ex_record.state != V10CredentialExchange.STATE_REQUEST_RECEIVED:
            raise CredentialManagerError(
                f"Credential exchange {cred_ex_record.credential_exchange_id} "
                f"in {cred_ex_record.state} state "
                f"(must be {V10CredentialExchange.STATE_REQUEST_RECEIVED})"
            )

        schema_id = cred_ex_record.schema_id
        rev_reg = None

        if cred_ex_record.credential:
            LOGGER.warning(
                "issue_credential called multiple times for "
                + "credential exchange record %s - abstaining",
                cred_ex_record.credential_exchange_id,
            )
        else:
            credential_offer = cred_ex_record.credential_offer
            credential_request = cred_ex_record.credential_request

            ledger = self._profile.inject(BaseLedger)
            async with ledger:
                schema = await ledger.get_schema(schema_id)
                credential_definition = await ledger.get_credential_definition(
                    cred_ex_record.credential_definition_id
                )

            tails_path = None
            if credential_definition["value"].get("revocation"):
                revoc = IndyRevocation(self._profile)
                try:
                    active_rev_reg_rec = await revoc.get_active_issuer_rev_reg_record(
                        cred_ex_record.credential_definition_id
                    )
                    rev_reg = await active_rev_reg_rec.get_registry()
                    cred_ex_record.revoc_reg_id = active_rev_reg_rec.revoc_reg_id

                    tails_path = rev_reg.tails_local_path
                    await rev_reg.get_or_fetch_local_tails_path()

                except StorageNotFoundError:
                    async with self._profile.session() as session:
                        posted_rev_reg_recs = (
                            await IssuerRevRegRecord.query_by_cred_def_id(
                                session,
                                cred_ex_record.credential_definition_id,
                                state=IssuerRevRegRecord.STATE_POSTED,
                            )
                        )
                    if not posted_rev_reg_recs:
                        # Send next 2 rev regs, publish tails files in background
                        async with self._profile.session() as session:
                            old_rev_reg_recs = sorted(
                                await IssuerRevRegRecord.query_by_cred_def_id(
                                    session,
                                    cred_ex_record.credential_definition_id,
                                )
                            )  # prefer to reuse prior rev reg size
                        for _ in range(2):
                            pending_rev_reg_rec = await revoc.init_issuer_registry(
                                cred_ex_record.credential_definition_id,
                                max_cred_num=(
                                    old_rev_reg_recs[0].max_cred_num
                                    if old_rev_reg_recs
                                    else None
                                ),
                            )
                            asyncio.ensure_future(
                                pending_rev_reg_rec.stage_pending_registry(
                                    self._profile,
                                    max_attempts=3,  # fail both in < 2s at worst
                                )
                            )
                    if retries > 0:
                        LOGGER.info(
                            "Waiting 2s on posted rev reg for cred def %s, retrying",
                            cred_ex_record.credential_definition_id,
                        )
                        await asyncio.sleep(2)
                        return await self.issue_credential(
                            cred_ex_record=cred_ex_record,
                            comment=comment,
                            retries=retries - 1,
                        )

                    raise CredentialManagerError(
                        f"Cred def id {cred_ex_record.credential_definition_id} "
                        "has no active revocation registry"
                    )
                del revoc

            credential_values = CredentialProposal.deserialize(
                cred_ex_record.credential_proposal_dict
            ).credential_proposal.attr_dict(decode=False)
            issuer = self._profile.inject(IndyIssuer)
            try:
                (
                    credential_json,
                    cred_ex_record.revocation_id,
                ) = await issuer.create_credential(
                    schema,
                    credential_offer,
                    credential_request,
                    credential_values,
                    cred_ex_record.credential_exchange_id,
                    cred_ex_record.revoc_reg_id,
                    tails_path,
                )

                # If the rev reg is now full
                if rev_reg and rev_reg.max_creds == int(cred_ex_record.revocation_id):
                    async with self._profile.session() as session:
                        await active_rev_reg_rec.set_state(
                            session,
                            IssuerRevRegRecord.STATE_FULL,
                        )

                    # Send next 1 rev reg, publish tails file in background
                    revoc = IndyRevocation(self._profile)
                    pending_rev_reg_rec = await revoc.init_issuer_registry(
                        active_rev_reg_rec.cred_def_id,
                        max_cred_num=active_rev_reg_rec.max_cred_num,
                    )
                    asyncio.ensure_future(
                        pending_rev_reg_rec.stage_pending_registry(
                            self._profile,
                            max_attempts=16,
                        )
                    )

            except IndyIssuerRevocationRegistryFullError:
                # unlucky: duelling instance issued last cred near same time as us
                async with self._profile.session() as session:
                    await active_rev_reg_rec.set_state(
                        session,
                        IssuerRevRegRecord.STATE_FULL,
                    )

                if retries > 0:
                    # use next rev reg; at worst, lucky instance is putting one up
                    LOGGER.info(
                        "Waiting 1s and retrying: revocation registry %s is full",
                        active_rev_reg_rec.revoc_reg_id,
                    )
                    await asyncio.sleep(1)
                    return await self.issue_credential(
                        cred_ex_record=cred_ex_record,
                        comment=comment,
                        retries=retries - 1,
                    )

                raise

            cred_ex_record.credential = json.loads(credential_json)

        cred_ex_record.state = V10CredentialExchange.STATE_ISSUED
        async with self._profile.session() as session:
            # FIXME - re-fetch record to check state, apply transactional update
            await cred_ex_record.save(session, reason="issue credential")

        credential_message = CredentialIssue(
            comment=comment,
            credentials_attach=[
                CredentialIssue.wrap_indy_credential(cred_ex_record.credential)
            ],
        )
        credential_message._thread = {"thid": cred_ex_record.thread_id}
        credential_message.assign_trace_decorator(
            self._profile.settings, cred_ex_record.trace
        )

        return (cred_ex_record, credential_message)

    async def receive_credential(
        self, message: CredentialIssue, connection_id: str
    ) -> V10CredentialExchange:
        """
        Receive a credential from an issuer.

        Hold in storage potentially to be processed by controller before storing.

        Returns:
            Credential exchange record, retrieved and updated

        """
        assert len(message.credentials_attach or []) == 1
        raw_credential = message.indy_credential(0)

        # FIXME use transaction, fetch for_update
        async with self._profile.session() as session:
            cred_ex_record = await (
                V10CredentialExchange.retrieve_by_connection_and_thread(
                    session,
                    connection_id,
                    message._thread_id,
                )
            )

            cred_ex_record.raw_credential = raw_credential
            cred_ex_record.state = V10CredentialExchange.STATE_CREDENTIAL_RECEIVED

            await cred_ex_record.save(session, reason="receive credential")
        return cred_ex_record

    async def store_credential(
        self, cred_ex_record: V10CredentialExchange, credential_id: str = None
    ) -> Tuple[V10CredentialExchange, CredentialAck]:
        """
        Store a credential in holder wallet; send ack to issuer.

        Args:
            cred_ex_record: credential exchange record
                with credential to store and ack
            credential_id: optional credential identifier to override default on storage

        Returns:
            Tuple: (Updated credential exchange record, credential ack message)

        """
        if cred_ex_record.state != (V10CredentialExchange.STATE_CREDENTIAL_RECEIVED):
            raise CredentialManagerError(
                f"Credential exchange {cred_ex_record.credential_exchange_id} "
                f"in {cred_ex_record.state} state "
                f"(must be {V10CredentialExchange.STATE_CREDENTIAL_RECEIVED})"
            )

        raw_credential = cred_ex_record.raw_credential
        revoc_reg_def = None
        ledger = self._profile.inject(BaseLedger)
        async with ledger:
            credential_definition = await ledger.get_credential_definition(
                raw_credential["cred_def_id"]
            )
            if (
                "rev_reg_id" in raw_credential
                and raw_credential["rev_reg_id"] is not None
            ):
                revoc_reg_def = await ledger.get_revoc_reg_def(
                    raw_credential["rev_reg_id"]
                )

        holder = self._profile.inject(IndyHolder)
        if (
            cred_ex_record.credential_proposal_dict
            and "credential_proposal" in cred_ex_record.credential_proposal_dict
        ):
            mime_types = CredentialPreview.deserialize(
                cred_ex_record.credential_proposal_dict["credential_proposal"]
            ).mime_types()
        else:
            mime_types = None

        if revoc_reg_def:
            revoc_reg = RevocationRegistry.from_definition(revoc_reg_def, True)
            await revoc_reg.get_or_fetch_local_tails_path()
        try:
            credential_id = await holder.store_credential(
                credential_definition,
                raw_credential,
                cred_ex_record.credential_request_metadata,
                mime_types,
                credential_id=credential_id,
                rev_reg_def=revoc_reg_def,
            )
        except IndyHolderError as e:
            LOGGER.error(f"Error storing credential. {e.error_code}: {e.message}")
            raise e

        credential_json = await holder.get_credential(credential_id)
        credential = json.loads(credential_json)

        cred_ex_record.state = V10CredentialExchange.STATE_ACKED
        cred_ex_record.credential_id = credential_id
        cred_ex_record.credential = credential
        cred_ex_record.revoc_reg_id = credential.get("rev_reg_id", None)
        cred_ex_record.revocation_id = credential.get("cred_rev_id", None)

        async with self._profile.session() as session:
            # FIXME - re-fetch record to check state, apply transactional update
            await cred_ex_record.save(session, reason="store credential")

        credential_ack_message = CredentialAck()
        credential_ack_message.assign_thread_id(
            cred_ex_record.thread_id, cred_ex_record.parent_thread_id
        )
        credential_ack_message.assign_trace_decorator(
            self._profile.settings, cred_ex_record.trace
        )

        if cred_ex_record.auto_remove:
            async with self._profile.session() as session:
                await cred_ex_record.delete_record(session)  # all done: delete

        return (cred_ex_record, credential_ack_message)

    async def receive_credential_ack(
        self, message: CredentialAck, connection_id: str
    ) -> V10CredentialExchange:
        """
        Receive credential ack from holder.

        Returns:
            credential exchange record, retrieved and updated

        """
        # FIXME use transaction, fetch for_update
        async with self._profile.session() as session:
            cred_ex_record = await (
                V10CredentialExchange.retrieve_by_connection_and_thread(
                    session,
                    connection_id,
                    message._thread_id,
                )
            )

            cred_ex_record.state = V10CredentialExchange.STATE_ACKED
            await cred_ex_record.save(session, reason="credential acked")

        if cred_ex_record.auto_remove:
            async with self._profile.session() as session:
                await cred_ex_record.delete_record(session)  # all done: delete

        return cred_ex_record
